import logging
from ibmsecurity.utilities import tools
from ibmsecurity.isam.aac.authentication import mechanism_types
from ibmsecurity.isam.aac.server_connections import smtp
from ibmsecurity.isam.aac.server_connections import ci
from ibmsecurity.isam.aac.server_connections import ws

logger = logging.getLogger(__name__)

# "uri" variable already used so using a different name
module_uri = "/iam/access/v8/authentication/mechanisms"
requires_modules = ["mga"]
requires_version = "8.0.0.0"


def get_all(isamAppliance, start=None, count=None, filter=None, sortBy=None, check_mode=False, force=False):
    """
    Retrieve a list of authentication mechanisms
    """
    return isamAppliance.invoke_get("Retrieve a list of authentication mechanisms",
                                    "{0}{1}".format(module_uri,
                                                    tools.create_query_string(start=start, count=count, filter=filter,
                                                                              sortBy=sortBy)),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieve a specific authentication mechanism
    """
    ret_obj = search(isamAppliance, name=name, check_mode=check_mode, force=force)
    id = ret_obj['data']

    if id == {}:
        logger.info(f"Authentication Mechanism {name} had no match, skipping retrieval.")
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieve a specific authentication mechanism",
                                        f"{module_uri}/{id}",
                                        requires_modules=requires_modules, requires_version=requires_version)


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Search authentication mechanism ID by name
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info(f"Found Authentication Mechanism {name} id: {obj['id']}")
            return_obj['data'] = obj['id']
            return_obj['rc'] = 0

    return return_obj


def set(isamAppliance, name, uri, description=None, attributes=None, properties=None, predefined=None,
        typeName=None, new_name=None, check_mode=False,
        force=False):
    """
    Creating or Modifying an Authentication Mechanism
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know authentication mechanism does not exist
        logger.info(f"Authentication Mechanism {name} had no match, requesting to add new one.")
        return add(isamAppliance, name, uri, description, attributes, properties, predefined, typeName, check_mode,
                   True)
    else:
        # Update request
        logger.info(f"Authentication Mechanism {name} exists, requesting to update.")
        return update(isamAppliance, name, uri, description, attributes, properties, predefined, typeName, new_name,
                      check_mode, force)


def add(isamAppliance, name, uri, description="", attributes=None, properties=None, predefined=False,
        typeName=None, check_mode=False, force=False):
    """
    Create a new Authentication Mechanism
    """
    if force is False:
        ret_obj = search(isamAppliance, name)

    if force is True or ret_obj['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            ret_obj = mechanism_types.search(isamAppliance, typeName)
            if ret_obj['data'] == {}:
                from ibmsecurity.appliance.ibmappliance import IBMError
                raise IBMError("999", f"Unable to find Authentication Mechanism Type: {typeName}")
            else:
                typeId = ret_obj['data']
            json_data = {
                "name": name,
                "uri": uri,
                "description": description,
                "predefined": predefined,
                "typeId": typeId
            }
            if attributes is not None:
                json_data['attributes'] = attributes
            if properties is not None:
                logger.info("Searching for keys to substitute value with uuids")
                for property in properties:
                    id = {}
                    if property['key'] == "EmailMessage.serverConnection":
                        id = smtp.search(isamAppliance, property['value'])['data']
                        logger.info(f"Found EmailMessage.serverConnection by name[{property['value']}] with uuid[{id}]")
                    elif property['key'] == "ScimConfig.serverConnection":
                        id = ws.search(isamAppliance, property['value'])['data']
                        logger.info(f"Found ScimConfig.serverConnection by name[{property['value']}] with uuid[{id}]")
                    elif property['key'] == "CI.serverConnection":
                        id = ci.search(isamAppliance, property['value'])['data']
                        logger.info(f"Found CI.serverConnection by name[{property['value']}] with uuid[{id}]")
                    if id != {}:
                        property['value'] = id
                json_data['properties'] = properties
            return isamAppliance.invoke_post(
                "Create a new Authentication Mechanism", module_uri, json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Delete an Authentication Mechanism
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    mech_id = ret_obj['data']

    if mech_id == {}:
        logger.info(f"Authentication Mechanism {name} not found, skipping delete.")
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete an Authentication Mechanism",
                f"{module_uri}/{mech_id}",
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, uri, description=None, attributes=None, properties=None, predefined=None,
           typeName=None, new_name=None, check_mode=False, force=False):
    """
    Update a specified authentication mechanism
    """
    mech_id, update_required, json_data = _check(isamAppliance, name, description, attributes, properties, predefined,
                                                 uri, typeName, new_name)
    if mech_id is None:
        from ibmsecurity.appliance.ibmappliance import IBMError
        raise IBMError("999", f"Cannot update data for unknown authentication mechanism: {name}")

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specified authentication mechanism",
                f"{module_uri}/{mech_id}", json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def _check(isamAppliance, name, description, attributes, properties, predefined, uri, typeName, new_name):
    """
    Check and return True if update needed

    TODO: Need to check for updated by only checking the attribute and properties that are provided (dont compare all)
    """
    update_required = False
    json_data = {
        "uri": uri
    }
    ret_obj = get(isamAppliance, name)
    if ret_obj['data'] == {}:
        logger.info("Authentication Mechanism not found, returning no update required.")
        return None, update_required, json_data
    else:
        mech_id = ret_obj['data']['id']
        if new_name is not None:
            json_data['name'] = new_name
        else:
            json_data['name'] = name
        if typeName is not None:
            ret_obj_type = mechanism_types.search(isamAppliance, typeName)
            if ret_obj_type['data'] == {}:
                from ibmsecurity.appliance.ibmappliance import IBMError
                raise IBMError("999", f"Unable to find Authentication Mechanism Type: {typeName}")
            else:
                json_data['typeId'] = ret_obj_type['data']
        else:
            del ret_obj['data']['typeId']
        if description is not None:
            json_data['description'] = description
        else:
            del ret_obj['data']['description']
        if predefined is not None:
            json_data['predefined'] = predefined
        else:
            del ret_obj['data']['predefined']
        if attributes is not None:
            json_data['attributes'] = attributes
        else:
            # May not exist so skip any exceptions when deleting
            try:
                del ret_obj['data']['attributes']
            except:
                pass
        if properties is not None:
            logger.info("Searching for keys to substitute value with uuids")
            for property in properties:
                id = {}
                if property['key'] == "EmailMessage.serverConnection":
                    id = smtp.search(isamAppliance, property['value'])['data']
                    logger.info(f"Found EmailMessage.serverConnection by name[{property['value']}] with uuid[{id}]")
                elif property['key'] == "ScimConfig.serverConnection":
                    id = ws.search(isamAppliance, property['value'])['data']
                    logger.info(f"Found ScimConfig.serverConnection by name[{property['value']}] with uuid[{id}]")
                elif property['key'] == "CI.serverConnection":
                    id = ci.search(isamAppliance, property['value'])['data']
                    logger.info(f"Found CI.serverConnection by name[{property['value']}] with uuid[{id}]")
                if id != {}:
                    property['value'] = id
            json_data['properties'] = properties
        else:
            # May not exist so skip any exceptions when deleting
            try:
                del ret_obj['data']['properties']
            except:
                pass
        del ret_obj['data']['id']
        import ibmsecurity.utilities.tools
        sorted_json_data = ibmsecurity.utilities.tools.json_sort(json_data)
        logger.debug(f"Sorted input: {sorted_json_data}")
        sorted_ret_obj = ibmsecurity.utilities.tools.json_sort(ret_obj['data'])
        logger.debug(f"Sorted existing data: {sorted_ret_obj}")
        if sorted_ret_obj != sorted_json_data:
            logger.info("Changes detected, update needed.")
            update_required = True

    return mech_id, update_required, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Authentication mechanisms between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['id']
        del obj['typeId']
    for obj in ret_obj2['data']:
        del obj['id']
        del obj['typeId']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['id', 'typeId'])



import logging

logger = logging.getLogger(__name__)

uri = "/isam/cluster/hvdb"

requires_model="Appliance"

def export(isamAppliance, filename, type='db2', check_mode=False, force=False):
    """
    Export the Runtime database
    """
    import os.path

    if force is True or os.path.exists(filename) is False:
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file("Export the Runtime database", f"{uri}/v1?type={type}",
                                                 filename, requires_model=requires_model)

    return isamAppliance.create_return_object()



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get existing activations
    """
    return isamAppliance.invoke_get("Retrieving activations",
                                    "/isam/capabilities/v1")


def get(isamAppliance, id=None, check_mode=False, force=False):
    """
    Retrieve a specified activation offering
    """
    if force is True or check(isamAppliance, id) is True:
        return isamAppliance.invoke_get("Retrieve a specified activation offering",
                                        f"/isam/capabilities/{id}/v1")

    return isamAppliance.create_return_object()


def update(isamAppliance, enabled, id=None, check_mode=False, force=False):
    """
    Update an activation offering
    """

    if isinstance(enabled, bool):
        if enabled is True:
            enabled = 'True'
        if enabled is False:
            enabled = 'False'

    if force is True or check_enabled(isamAppliance, id, enabled) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Update an activation offering",
                                            f"/isam/capabilities/{id}/v1",
                                            {'enabled': enabled.capitalize()}
                                           )

    return isamAppliance.create_return_object()


def set(isamAppliance, id, code, check_mode=False, force=False):
    """
    Activate an ISAM module
    """
    if force is True or check(isamAppliance, id) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            ret_obj = isamAppliance.invoke_post(
                "Activating a Module",
                "/isam/capabilities/v1",
                {
                    'code': code
                })
            # Update 'facts', with newly activated module
            if 'activations' not in isamAppliance.facts:
                isamAppliance.facts['activations'] = []
            isamAppliance.facts['activations'].append(id)
            return ret_obj

    return isamAppliance.create_return_object()


def delete(isamAppliance, id, check_mode=False, force=False):
    """
    Delete activation of an ISAM module
    """
    if force is True or check(isamAppliance, id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            ret_obj = isamAppliance.invoke_delete(
                "Deleting activation of Module",
                f"/isam/capabilities/{id}/v1")
            # Update 'facts', remove module
            isamAppliance.facts['activations'].remove(id)
            return ret_obj

    return isamAppliance.create_return_object()


def check(isamAppliance, id):
    """
    Check if ISAM module is already activated
    """
    ret_obj = get_all(isamAppliance)

    for activation in ret_obj['data']:
        if id == activation['id']:
            logger.info("This module is already activated: " + id)
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare activations of ISAM modules between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])


def check_enabled(isamAppliance, id, enabled):
    """
    Check to see if the enabled is the same as the value passed in
    """
    ret_obj = get_all(isamAppliance)

    for activation in ret_obj['data']:
        if id == activation['id']:
            if activation['enabled'] == enabled.capitalize():
                return True

    return False



import logging
import ibmsecurity.utilities.tools
import ibmsecurity.isam.base.network.interfaces

logger = logging.getLogger(__name__)
requires_model="Appliance"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the DNS configuration
    """
    return isamAppliance.invoke_get("Retrieving the DNS configuration",
                                    "/net/dns", requires_model=requires_model)


def set(isamAppliance, primaryServer=None, secondaryServer=None, tertiaryServer=None, searchDomains=None, auto=True,
        autoFromInterface=None, check_mode=False, force=False):
    """
    Updating the DNS configuration
    """

    if isinstance(auto, basestring):
        if auto.lower() == 'true':
            auto = True
        else:
            auto = False

    # check autoFromInterface. If it is a label replace it with corresponding interface UUID, else leave it untouched (treat as UUID)
    if autoFromInterface is not None:
      ret_obj = ibmsecurity.isam.base.network.interfaces.get_all(isamAppliance)
      for intfc in ret_obj['data']['interfaces']:
        if intfc['label'] == autoFromInterface:
          autoFromInterface = intfc['uuid']

    check_value,warnings = _check(isamAppliance, primaryServer, secondaryServer, tertiaryServer, searchDomains, auto,
                               autoFromInterface)
    if force is True or check_value is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Updating the DNS configuration",
                "/net/dns",
                {
                    'auto': auto,
                    'autoFromInterface': autoFromInterface,
                    'primaryServer': primaryServer,
                    'secondaryServer': secondaryServer,
                    'tertiaryServer': tertiaryServer,
                    'searchDomains': searchDomains
                }, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def test(isamAppliance, host, server=None, force=False, check_mode=False):
    """
    Run DNS Lookup Test
    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)

    ret_obj = isamAppliance.invoke_post("Run DNS Lookup Test",
                                        "/isam/net/lookup",
                                        {
                                            'host': host,
                                            'server': server
                                        }, requires_model=requires_model)
    # HTTP POST calls get flagged as changes - but DNS lookup changes nothing so override
    if ret_obj['changed'] is True:
        ret_obj['changed'] = False

    return ret_obj


def _check(isamAppliance, primaryServer, secondaryServer, tertiaryServer, searchDomains, auto=False,
           autoFromInterface=None):
    """
    Check if DNS is already set
    """

    ret_obj = get(isamAppliance)
    check_value,warnings=True,ret_obj['warnings']

    check_json_data = {
        'auto': auto,
        'autoFromInterface': autoFromInterface,
        'primaryServer': primaryServer,
        'secondaryServer': secondaryServer,
        'tertiaryServer': tertiaryServer,
        'searchDomains': searchDomains
    }

    if ibmsecurity.utilities.tools.json_sort(ret_obj['data']) == ibmsecurity.utilities.tools.json_sort(check_json_data):
        check_value=True
        return check_value,warnings
    else:
        check_value = False
        return check_value,warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare dns settings between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/databases/hvdb"
requires_modules = None
requires_version = None


def get(isamAppliance, check_mode=False, force=False):
    """
    Get the status of the most recent deletion of data that completed

    """
    return isamAppliance.invoke_get("Get the status of the most recent deletion of data that completed",
                                    f"{uri}/status",
                                    requires_modules=requires_modules, requires_version=requires_version)


def cancel_all(isamAppliance, check_mode=False, force=False):
    """
    Cancel the deletion of all data

    """
    return isamAppliance.invoke_get("Cancel the deletion of all data",
                                    f"{uri}/cancel",
                                    requires_modules=requires_modules, requires_version=requires_version)


def cancel_all_user(isamAppliance, check_mode=False, force=False):
    """
    Cancel the deletion of all user data

    """
    return isamAppliance.invoke_get("Cancel the deletion of all user data",
                                    f"{uri}/userdata/cancel",
                                    requires_modules=requires_modules, requires_version=requires_version)


def cancel_all_devices(isamAppliance, check_mode=False, force=False):
    """
    Cancel the deletion of all user devices

    """
    return isamAppliance.invoke_get("Cancel the deletion of all user devices",
                                    f"{uri}/devices/cancel",
                                    requires_modules=requires_modules, requires_version=requires_version)


def clear(isamAppliance, check_mode=False, force=False):
    """
    Clear the status of the most recent deletion of data that completed

    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:

        return isamAppliance.invoke_delete(
            "Clear the status of the most recent deletion of data that completed",
            f"{uri}/status",
            requires_modules=requires_modules, requires_version=requires_version
        )

    return isamAppliance.create_return_object()


def delete_all(isamAppliance, check_mode=False, force=False):
    """
    Delete all data

    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:

        return isamAppliance.invoke_delete(
            "Delete all data",
            f"{uri}",
            requires_modules=requires_modules, requires_version=requires_version
        )

    return isamAppliance.create_return_object()


def delete_user(isamAppliance, userId, check_mode=False, force=False):
    """
    Delete all user data for a specific user

    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:

        return isamAppliance.invoke_delete(
            "Delete all user data for a specific user",
            f"{uri}/{userId}",
            requires_modules=requires_modules, requires_version=requires_version
        )

    return isamAppliance.create_return_object()


def delete_all_user(isamAppliance, check_mode=False, force=False):
    """
    Delete all user data
    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:

        return isamAppliance.invoke_delete(
            "Delete all user data",
            f"{uri}/userdata",
            requires_modules=requires_modules, requires_version=requires_version
        )

    return isamAppliance.create_return_object()


def delete_all_devicces(isamAppliance, check_mode=False, force=False):
    """
    Delete all user devices
    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:

        return isamAppliance.invoke_delete(
            "Delete all user devices",
            f"{uri}/devices",
            requires_modules=requires_modules, requires_version=requires_version
        )

    return isamAppliance.create_return_object()



import logging
import ibmsecurity.isam.base.management_authorization.role

logger = logging.getLogger(__name__)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieving the list of users for an authorization roles
    """
    return isamAppliance.invoke_get("Retrieving the list of users for an authorization roles",
                                    f"/authorization/roles/{name}/users/v1")


def set(isamAppliance, name, user_name, type='embedded_ldap', check_mode=False, force=False):
    """
    Add a user to management authorization role
    """
    new_user = True
    ret_obj = ibmsecurity.isam.base.management_authorization.role.get(isamAppliance, name)

    if (ret_obj['data']['users'] == None):
        ret_obj['data']['users'] = []
    else:
        for usr in ret_obj['data']['users']:
            if usr['name'] == user_name:
                if usr['type'] == type:
                    if force is False:
                        return isamAppliance.create_return_object()
                    new_user = False
                else:  # Replace user with new type
                    ret_obj['data']['users'].remove(usr)
                break

    if new_user is True:
        ret_obj['data']['users'].append({'name': user_name, 'type': type})

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Add user to management authorization role",
            f"/authorization/roles/{name}/v1", ret_obj['data'])


def delete(isamAppliance, name, user_name, check_mode=False, force=False):
    """
    Delete a user from management authorization role
    """
    user_found = False
    ret_obj = ibmsecurity.isam.base.management_authorization.role.get(isamAppliance, name)

    if (ret_obj['data']['users'] != None):
        for usr in ret_obj['data']['users']:
            if usr['name'] == user_name:
                user_found = True
                ret_obj['data']['users'].remove(usr)
                break

    if user_found is False and force is False:
        return isamAppliance.create_return_object()

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put(
            "Delete user from management authorization role",
            f"/authorization/roles/{name}/v1", ret_obj['data'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Get all snmp objects
    """
    return isamAppliance.invoke_get("Get all alert objects",
                                    "/core/system_alerts")


def enable(isamAppliance, uuid, objType, check_mode=False, force=False):
    """
    Enable a system alert
    """
    if force is True or _check(isamAppliance, uuid) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Enable a system alert",
                "/core/system_alerts",
                {
                    'uuid': uuid,
                    'objType': objType
                })

    return isamAppliance.create_return_object()


def disable(isamAppliance, uuid, check_mode=False, force=False):
    """
    Delete a system alert
    """
    if force is True or _check(isamAppliance, uuid) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a system alert",
                f"/core/system_alerts/{uuid}")

    return isamAppliance.create_return_object()


def _check(isamAppliance, uuid):
    """
    Check if the system alert exists or not
    """
    ret_obj = get(isamAppliance)
    for obj in ret_obj['data']['responses']:
        if obj['uuid'] == uuid:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare system alert objects between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    for obj in ret_obj1['data']['responses']:
        del obj['uuid']
    for obj in ret_obj2['data']['responses']:
        del obj['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])



import logging
import os.path

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the general configuration
    """
    return isamAppliance.invoke_get("Retrieving the general configuration", "/isam/packet_tracing",
                                    requires_model=requires_model)


def execute(isamAppliance, enabled, filter=None, interface=None, max_size=None, check_mode=False,
            force=False, snaplen=None):
    """
    Execute an operation (start, stop or restart) on packet tracing
    """

    check_value, warnings = _check(isamAppliance, enabled)
    json_data = {"enable": enabled}

    if filter is not None:
        json_data['filter'] = filter

    if interface is not None:
        json_data['interface'] = interface

    if max_size is not None:
        json_data['max_size'] = interface

    if snaplen is not None:
        if tools.version_compare(isamAppliance.facts['version'], "9.0.6.0") < 0:
            warnings.append(
                f"Appliance at version: {isamAppliance.facts['version']}, snaplen: {snaplen} is not supported. Needs 9.0.6.0 or higher.  Ignoring snaplen for this call."
            )
        else:
            json_data["snaplen"] = snaplen

    if force is True or check_value is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Executing an operation on packet trace", "/isam/packet_tracing/", json_data,
                requires_model=requires_model, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, check_mode=False, force=False):
    """
    Clearing the packet tracing PCAP files
    """

    ret_obj = get(isamAppliance)

    if force is True or ret_obj['data']['files'] != []:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=ret_obj['warnings'])
        else:
            return isamAppliance.invoke_delete(
                "Clearing the packet tracing PCAP files", "/isam/packet_tracing/", requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=ret_obj['warnings'])


def export_file(isamAppliance, filepath, filename, check_mode=False, force=False):
    """
    Exporting the packet tracing PCAP file
    """
    if force is True or os.path.exists(filepath) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_get_file(
                "Exporting the packet tracing PCAP file",
                f"/isam/packet_tracing/pcap/{filename}?export", filepath, requires_model=requires_model
            )

    return isamAppliance.create_return_object()


def _check(isamAppliance, enabled):
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']

    if 'enabled' in ret_obj['data']:
        return ret_obj['data']['enabled'] == enabled, warnings
    else:
        return True, warnings



import logging
import ibmsecurity.utilities.tools
import os.path
import shutil
import zipfile
from ibmsecurity.isam.web.reverse_proxy.management_root import directory
from ibmsecurity.isam.web.reverse_proxy.management_root import file
from ibmsecurity.isam.web.reverse_proxy import instance
from ibmsecurity.utilities.tools import get_random_temp_dir, files_same_zip_content

logger = logging.getLogger(__name__)


def export_zip(isamAppliance, instance_id, filename, check_mode=False, force=False):
    """
    Exporting the contents of the administration pages root as a .zip file

    :param isamAppliance:
    :param instance_id:
    :param filename:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or os.path.exists(filename) is False:
        if check_mode is False:
            return isamAppliance.invoke_get_file(
                "Exporting the contents of the administration pages root as a .zip file",
                "/wga/reverseproxy/{0}/management_root?index=&name=&enc_name=&type=&browser=".format(instance_id),
                filename=filename, no_headers=True)

    return isamAppliance.create_return_object()


def import_zip(isamAppliance, instance_id, filename, delete_missing=False, check_mode=False, force=False):
    """
    Importing the contents of a .zip file to the administration pages root
    Feature delete_missing will compare import zip with server content and delete missing files in the import zip from server
    """
    warnings = []

    if force is True or _check_import(isamAppliance, instance_id, filename):
        if delete_missing is True:
            tempdir = get_random_temp_dir()
            tempfilename = "management_root.zip"
            tempfile =  os.path.join(tempdir, tempfilename)
            export_zip(isamAppliance, instance_id, tempfile)

            zServerFile = zipfile.ZipFile(tempfile)
            zClientFile = zipfile.ZipFile(filename)

            files_on_server = [];
            for info in zServerFile.infolist():
                files_on_server.append(info.filename)
            files_on_client = [];
            for info in zClientFile.infolist():
                files_on_client.append(info.filename)
            missing_client_files = [x for x in files_on_server if x not in files_on_client]

            if missing_client_files != []:
                logger.info("list all missing files in {}, which will be deleted on the server: {}.".format(filename, missing_client_files))

            for x in missing_client_files:
                if x.endswith('/'):
                    search_dir= os.path.dirname(x[:-1]) + '/'
                    if search_dir not in missing_client_files:
                        logger.debug("delete directory on the server: {0}.".format(x))
                        directory.delete(isamAppliance, instance_id, x, check_mode=check_mode)
                else:
                    search_dir= os.path.dirname(x) + '/'
                    if search_dir not in missing_client_files:
                        logger.debug("delete file on the server: {0}.".format(x))
                        file.delete(isamAppliance, instance_id, x, check_mode=check_mode)
            shutil.rmtree(tempdir)

        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing the contents of a .zip file to the administration pages root",
                 "/wga/reverseproxy/{0}/management_root".format(instance_id),
                [
                    {
                        'file_formfield': 'file',
                        'filename': filename,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {
                    "force": force
                }, json_response=False)

    return isamAppliance.create_return_object(warnings=warnings)

def _check_import(isamAppliance, instance_id, filename):
    """
    Checks if runtime template zip from server and client differ
    :param isamAppliance:
    :param filename:
    :return:
    """

    if not instance._check(isamAppliance, instance_id):
      logger.info("instance {} does not exist on this server. Skip import".format(instance_id))
      return False

    tempdir = get_random_temp_dir()
    tempfilename = "management_root.zip"
    tempfile =  os.path.join(tempdir, tempfilename)
    export_zip(isamAppliance, instance_id, tempfile)

    identical = files_same_zip_content(filename,tempfile)

    shutil.rmtree(tempdir)
    if identical:
        logger.info("management_root files {} are identical with the server content. No update necessary.".format(filename))
        return False
    else:
        logger.info("management_root files {} differ from the server content. Updating management_root files necessary.".format(filename))
        return True

def check(isamAppliance, instance_id, id, name, type, check_mode=False, force=False):
    ret_obj = None

    if (type.lower() == 'directory'):
        ret_obj = directory._check(isamAppliance, instance_id, id, name)
    elif (type.lower() == 'file'):
        ret_obj = file._check(isamAppliance, instance_id, id, name)
    else:
        type = 'unknown'

    data = {
        'id': ret_obj,
        'top': id,
        'name': name,
        'type': type
    }

    return isamAppliance.create_return_object(data=data)



import logging
import copy
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/isam/cluster/v2"
requires_modules = None
requires_version = None
requires_model = "Docker"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve the current database configuration
    """
    return isamAppliance.invoke_get("Retrieve the current database configuration", uri,
                                    requires_modules=requires_modules, requires_version=requires_version,
                                    requires_model=requires_model)


def set(isamAppliance,
        hvdb_db_type,
        hvdb_address,
        hvdb_port,
        hvdb_user,
        hvdb_password,
        hvdb_db_name=None,
        hvdb_db_secure=None,
        hvdb_driver_type=None,
        hvdb_db2_alt_address=None,
        hvdb_db2_alt_port=None,
        cfgdb_embedded=True,
        check_mode=False,
        force=False,
        ignore_password_for_idempotency=False,
        **kwargs):
    """
    Set cluster configuration
    hvdb_db_type	String	No	The type of database that is being used. Valid values are db2, postgresql, mssql and oracle.
    hvdb_address	String	No	The IP or hostname of the external database server.
    hvdb_port	    Integer	No	The port on which the external database server is listening.
    hvdb_user	    String	No	The administrator name for the external database.
    hvdb_password	String	No	The administrator password for the external database.
    hvdb_db2_alt_address	String	Yes	The IP or hostname of the fail-over server in DB2 HADR/ACR environments. (Can be specified for DB2 only).
    hvdb_db2_alt_port	    String	Yes	The IP or hostname of the fail-over server in DB2 HADR/ACR environments. (Can be specified for DB2 only).
    hvdb_db_name	String	No	The name of the external database (Required for DB2, PostgreSQL, MSSQL and Oracle).
    hvdb_db_secure	Boolean	No	A flag true/false indicating whether or not the external database is secure. (Required for DB2, PostgreSQL, MSSQL and Oracle).
    hvdb_db_truststore	String	Yes	The SSL Key Store which contains the trusted certificate of the Oracle DB requiring secure connectivity. (Required for Oracle only when hvdb_db_secure is true)
    hvdb_driver_type	String	Yes	The type of Oracle JDBC driver to use. (Required for Oracle only).
    hvdb_failover_servers	FailoverServer[]	Yes	An ordered list of PostgreSQL failover servers for the Runtime database. (Can be specified for PostgreSQL only).
         Note: This is an array of elements.

    - new in 10.0.0.8  configdb (embedded)

    cfgdb_embedded	Boolean	No	A flag true/false indicating whether or not the Configuration database is embedded (true) or external (false). If this value is not specified, the internal database will be used and none of the other cfgdb_* values are required.
    cfgdb_db_type	String	Yes	The type of database that is being used. Valid values are db2, postgresql, mssql and oracle. This option is only valid if cfgdb_embedded is set to false.
    cfgdb_address	String	Yes	The IP or hostname of the external database server. This option is only valid if cfgdb_embedded is set to false.
    cfgdb_port	Integer	Yes	The port on which the external database server is listening. This option is only valid if cfgdb_embedded is set to false.
    cfgdb_user	String	Yes	The administrator name for the external database. This option is only valid if cfgdb_embedded is set to false.
    cfgdb_password	String	Yes	The administrator password for the external database. This option is only valid if cfgdb_embedded is set to false.
    cfgdb_db2_alt_address	String	Yes	The IP or hostname of the fail-over server in DB2 HADR/ACR environments. (Can be specified for DB2 only). This option is only valid if cfgdb_embedded is set to false.
    cfgdb_db2_alt_port	String	Yes	The IP or hostname of the fail-over server in DB2 HADR/ACR environments. (Can be specified for DB2 only). This option is only valid if cfgdb_embedded is set to false.
    cfgdb_db_name	String	Yes	The name of the external database (Required for DB2, PostgreSQL, MSSQL and Oracle). This option is only valid if cfgdb_embedded is set to false.
    cfgdb_db_secure	Boolean	Yes	A flag true/false indicating whether or not the external database is secure. (Required for DB2, PostgreSQL, MSSQL and Oracle). This option is only valid if cfgdb_embedded is set to false.
    cfgdb_db_truststore	String	Yes	The SSL Key Store which contains the trusted certificate of the Oracle DB requiring secure connectivity. (Required for Oracle only when cfgdb_db_secure is true)
    cfgdb_driver_type	String	Yes	The type of Oracle JDBC driver to use. (Required for Oracle only). This option is only valid if cfgdb_embedded is set to false.
    cfgdb_failover_servers	FailoverServer[]	Yes	An ordered list of PostgreSQL failover servers for the Configuration database. (Can be specified for PostgreSQL only).
    Note: This is an array of elements.

    - new in 11 - MSSQL

    hvdb_mssql_instance_name	String	Yes	The database instance name to connect to. When it isn't specified, a connection is made to the default instance. (Can be specified for MSSQL only)
    hvdb_mssql_trust_server_cert	Boolean	No	Set to "true" to specify that the driver doesn't validate the server TLS/SSL certificate. (Can be specified for MSSQL only)
    hvdb_mssql_failover_partner	String	Yes	The name of the failover server used in a database mirroring configuration. This property is used for an initial connection failure to the principal server. After you make the initial connection, this property is ignored. (Can be specified for MSSQL only)

    cfgdb_mssql_instance_name	String	Yes	The database instance name to connect to. When it isn't specified, a connection is made to the default instance. (Can be specified for MSSQL only)
    cfgdb_mssql_trust_server_cert	Boolean	No	Set to "true" to specify that the driver doesn't validate the server TLS/SSL certificate. (Can be specified for MSSQL only)
    cfgdb_mssql_failover_partner	String	Yes	The name of the failover server used in a database mirroring configuration. This property is used for an initial connection failure to the principal server. After you make the initial connection, this property is ignored. (Can be specified for MSSQL only)
    """
    warnings = []
    if hvdb_db_name is None:
        warnings.append(f"hvdb_db_type {hvdb_db_type} requires hvdb_db_name")
        return isamAppliance.create_return_object(warnings=warnings)
    if hvdb_db_secure is None:
        warnings.append(f"hvdb_db_type {hvdb_db_type} requires hvdb_db_secure")
        return isamAppliance.create_return_object(warnings=warnings)
    if hvdb_db_type == "oracle":
        if hvdb_driver_type is None:
            warnings.append(f"hvdb_db_type {hvdb_db_type} requires hvdb_driver_type")
            return isamAppliance.create_return_object(warnings=warnings)
        if hvdb_db_secure and kwargs.get("hvdb_db_truststore", None) is None:
            warnings.append(f"hvdb_db_type {hvdb_db_type} with hvdb_db_secure requires hvdb_db_truststore")
            return isamAppliance.create_return_object(warnings=warnings)
    if isinstance(hvdb_port, basestring):
        hvdb_port = int(hvdb_port)

    # Create a simple json with just the main client attributes
    db_json = {
        "hvdb_db_type": hvdb_db_type,
        "hvdb_address": hvdb_address,
        "hvdb_port": hvdb_port,
        "hvdb_user": hvdb_user,
        "hvdb_password": hvdb_password
    }
    # Add attributes that have been supplied... otherwise skip them.
    if hvdb_db2_alt_address is not None:
        db_json["hvdb_db2_alt_address"] = hvdb_db2_alt_address
    if hvdb_db2_alt_port is not None:
        db_json["hvdb_db2_alt_port"] = hvdb_db2_alt_port
    if hvdb_db_name is not None:
        db_json["hvdb_db_name"] = hvdb_db_name
    if hvdb_db_secure is not None:
        db_json["hvdb_db_secure"] = hvdb_db_secure
    if hvdb_driver_type is not None:
        db_json["hvdb_driver_type"] = hvdb_driver_type

    if tools.version_compare(isamAppliance.facts['version'], "10.0.8.0") < 0:
        warnings.append(
                f"Appliance at version: {isamAppliance.facts['version']}, cfgdb_embedded is not supported. Needs 10.0.8.0 or higher. Ignoring for this call.")
    else:
        db_json["cfgdb_embedded"] = cfgdb_embedded

    for k, v in kwargs.items():
        if k.startswith("cfgdb_mssql_"):
            if tools.version_compare(isamAppliance.facts['version'], "10.0.9.0") < 0:
                warnings.append(
                    f"Appliance at version: {isamAppliance.facts['version']}, MSSQL parameters are not supported ({k}). Needs 10.0.8.0 or higher. Ignoring for this call.")
            else:
                db_json[k] = v
            continue
        if k.startswith("hvdb_mssql_"):
            if tools.version_compare(isamAppliance.facts['version'], "10.0.9.0") < 0:
                warnings.append(
                    f"Appliance at version: {isamAppliance.facts['version']}, MSSQL parameters are not supported ({k}). Needs 10.0.8.0 or higher. Ignoring for this call.")
            else:
                db_json[k] = v
            continue
        if k.startswith("cfgdb_"):
            if tools.version_compare(isamAppliance.facts['version'], "10.0.8.0") < 0:
                warnings.append(
                    f"Appliance at version: {isamAppliance.facts['version']}, cfgdb_* parameters are not supported ({k}). Needs 10.0.8.0 or higher. Ignoring for this call.")
            else:
                db_json[k] = v
            continue
        db_json[k] = v

    obj = _check(isamAppliance=isamAppliance, db_json=db_json,
                 ignore_password_for_idempotency=ignore_password_for_idempotency)

    if force or not obj['value']:
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=obj['warnings'])
        else:
            return isamAppliance.invoke_post("Set database configuration", uri, db_json,
                                             requires_modules=requires_modules, requires_version=requires_version,
                                             requires_model=requires_model, warnings=obj['warnings'])

    return isamAppliance.create_return_object(warnings=obj['warnings'])


def _check(isamAppliance, db_json, ignore_password_for_idempotency=False):
    """
    Check if provided json values match the configuration on appliance

    :param isamAppliance:
    :param db_json:
    :return:
    """

    obj = {'value': True, 'warnings': ""}

    ret_obj = get(isamAppliance)

    obj['warnings'] = ret_obj['warnings']
    del_password = False

    if ignore_password_for_idempotency is True:
        if 'hvdb_password' in ret_obj['data']:
            del ret_obj['data']['hvdb_password']
        if 'hvdb_password' in db_json:
            password = db_json['hvdb_password']
            del_password = True
            del db_json['hvdb_password']

    sorted_ret_obj = tools.json_sort(ret_obj['data'])
    sorted_json_data = tools.json_sort(db_json)
    logger.debug(f"Sorted Existing Data:{sorted_ret_obj}")
    logger.debug(f"Sorted Desired  Data:{sorted_json_data}")

    if del_password:
        db_json['hvdb_password'] = password

    if sorted_ret_obj != sorted_json_data:
        obj['value'] = False
        return obj
    else:
        obj['value'] = True
        return obj


def compare(isamAppliance1, isamAppliance2):
    """
    Compare cluster configuration between two appliances
    """

    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/v1/property/propertyfiles"


def get_all(isvgAppliance, check_mode=False, force=False):
    """
    Retrieve a list of property files
    """
    return isvgAppliance.invoke_get("Retrieve a list of property files", uri)


def _get(isvgAppliance, propertyFile):
    return isvgAppliance.invoke_get("Retrieve properties for a specific file", "{0}/{1}".format(uri, propertyFile))


def search(isvgAppliance, propertyFile, check_mode=False, force=False):
    """
    Search property file by name
    """
    ret_obj = get_all(isvgAppliance)
    return_obj = isvgAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['PropertyFile'] == propertyFile:
            logger.info("Found property file {0} type: {1}".format(propertyFile, obj['type']))
            return_obj['data'] = obj['PropertyFile']
            return_obj['rc'] = 0

    return return_obj



import logging

logger = logging.getLogger(__name__)
uri = "/sys"
requires_model = "Appliance"


def get_routes(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current active system routes
    """
    return isamAppliance.invoke_get("Retrieving the current active system routes", f"{uri}/routes",
                                    requires_model=requires_model)


def get_addresses(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the currently assigned addresses
    """
    return isamAppliance.invoke_get("Retrieving the currently assigned addresses", f"{uri}/ifaces",
                                    requires_model=requires_model)



import logging
import os.path
from ibmsecurity.utilities import tools
from io import open

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/fsso_config"
requires_modules = ["wga"]
requires_version = None


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a list of all FSSO configuration files
    """
    return isamAppliance.invoke_get("Retrieving a list of all FSSO configuration files",
                                    "{0}/".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, name, check_mode=False, force=False):
    """
    Retrieving an FSSO configuration file
    """

    ret_obj = search(isamAppliance, name, check_mode=False, force=False)
    id = ret_obj['data']

    if id == {}:
        warnings = ["Retrieving an FSSO configuration file " + name + " does not exist"]
        logger.info("Retrieving an FSSO configuration file " + name + " does not exist")
        return isamAppliance.create_return_object(warnings=warnings)

    else:
        return isamAppliance.invoke_get("Retrieving an FSSO configuration file",
                                        "{0}/{1}".format(uri, id),
                                        requires_modules=requires_modules,
                                        requires_version=requires_version
                                       )


def get_template(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the FSSO configuration file template
    """

    return isamAppliance.invoke_get("Retrieving the FSSO configuration file template",
                                    "/isam/wga_templates/fsso_template",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                   )


def add(isamAppliance, name, data, check_mode=False, force=False):
    """
    Creating an FSSO configuration file
    """
    ret_obj = search(isamAppliance, name, check_mode=False, force=False)
    id = ret_obj['data']

    if id != {}:
        logger.info("FSSO Configuration file '{0}' already exists.  Skipping add.".format(name))

    if force is True or id == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating an FSSO configuration file",
                "{0}".format(uri),
                {
                    'name': name,
                    'fsso_config_data': data
                },
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def update(isamAppliance, name, filepath=None, data=None, check_mode=False, force=False):
    """
    Updating an existing FSSO configuration file

    """

    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    contents = ""
    update_required = False

    if id != {}:
        ret_obj = get(isamAppliance, name)
        ret_data = ret_obj['data']['contents']

        if filepath != None:
            if os.path.exists(filepath) is True:
                openfile = open(filepath)
                filecontent = openfile.read()
                if filecontent != ret_data:
                    contents = filecontent
                    update_required = True
        elif data != None:
            if ret_data != data:
                contents = data
                update_required = True

    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_put(
                "Updating an existing FSSO configuration file",
                "{0}/{1}".format(uri, name),
                {
                    'id': name,
                    'fsso_config_data': contents
                },
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def set(isamAppliance, name, filepath='', data=None, check_mode=False, force=False):
    """
    Creating or Modifying a FSSO Configuration File
    """
    ret_obj = search(isamAppliance, name=name)
    if ret_obj['data'] == {}:
        # If no id was found, Force the add
        return add(isamAppliance, name, data=data, check_mode=check_mode, force=force)
    else:
        # Update PIP
        return update(isamAppliance, name, filepath=filepath, data=data, check_mode=check_mode, force=force)


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Deleting an FSSO configuration file
    """
    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    if id == {}:
        logger.info("FSSO configuration file '{0}' does not exists.  Skipping delete.".format(name))

    if force is True or id != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_delete(
                "Deleting an FSSO configuration file",
                "{0}/{1}".format(uri, id),
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def export_file(isamAppliance, name, filepath, check_mode=False, force=False):
    """
    Exporting an FSSO configuration file

    """

    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    if id == {}:
        logger.info("FSSO configuration file '{0}' does not exists.  Skipping export file.".format(name))
        warnings = ["FSSO configuration file '{0}' does not exists.  Skipping export file.".format(name)]
        return isamAppliance.create_return_object(warnings=warnings)

    if os.path.exists(filepath) is True:
        logger.info("File '{0}' already exists.  Skipping export.".format(filepath))
        warnings = ["File '{0}' already exists.  Skipping export.".format(filepath)]
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or id != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_get_file(
                "Exporting an FSSO configuration file",
                "{0}/{1}?export".format(uri, name), filepath,
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def import_file(isamAppliance, filepath, check_mode=False, force=False):
    """
    Importing a new FSSO configuration file
    """
    filename = os.path.basename(filepath)

    ret_obj = search(isamAppliance, name=filename)
    id = ret_obj['data']

    if os.path.exists(filepath) is False:
        logger.info("File '{0}' does not exists.  Skipping import.".format(filepath))
        warnings = ["File '{0}' does not exists.  Skipping import.".format(filepath)]
        return isamAppliance.create_return_object(warnings=warnings)

    if id != {}:
        logger.info("FSSO configuration file '{0}' already exists.  Skipping import file.".format(filename))

    if force is True or id == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Importing a new FSSO configuration file",
                "{0}/".format(uri),
                [
                    {
                        'file_formfield': 'fsso_config_file',
                        'filename': filepath,
                        'mimetype': 'application/octet-stream'
                    }
                ],
                {},
                requires_modules=requires_modules, requires_version=requires_version
            )
    return isamAppliance.create_return_object()


def rename(isamAppliance, name, new_name, check_mode=False, force=False):
    """
    Renaming a FSSO configuration file
    """

    ret_obj = search(isamAppliance, name=name)
    id = ret_obj['data']

    if id == {}:
        logger.info("FSSO configuration file '{0}' does not exists.  Skipping rename.".format(name))
        warnings = ["FSSO configuration file '{0}' does not exists.  Skipping rename.".format(name)]
        return isamAppliance.create_return_object(warnings=warnings)

    if force is True or id != {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:

            return isamAppliance.invoke_put(
                "Renaming a FSSO configuration file",
                "{0}/{1}".format(uri, name),
                {"new_name": new_name},
                requires_modules=requires_modules, requires_version=requires_version
            )

    return isamAppliance.create_return_object()


def compare(isamAppliance1, isamAppliance2):
    """
    Compare FSSO configuration files
    """

    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        ret_obj = get(isamAppliance1, obj['id'])['data']
        del obj['version']
        obj['contents'] = ret_obj['contents']
    for obj in ret_obj2['data']:
        ret_obj = get(isamAppliance2, obj['id'])['data']
        del obj['version']
        obj['contents'] = ret_obj['contents']

    return tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['version'])


def search(isamAppliance, name, force=False, check_mode=False):
    """
    Retrieve ID for named FSSO
    """
    ret_obj = get_all(isamAppliance)

    ret_obj_new = isamAppliance.create_return_object()

    for obj in ret_obj['data']:
        if obj['id'] == name:
            logger.info("Found FSSO '{0}' id: '{1}'".format(name, obj['id']))
            ret_obj_new['data'] = obj['id']

    return ret_obj_new



import logging

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve runtime component replicaiton status
    """
    return isamAppliance.invoke_get("Retrieving web runtime component replication status",
                                    "/isam/runtime_components?cluster=true")


def set(isamAppliance, replicating, check_mode=False, force=False):
    """
    Set runtime component replication status
    """
    if force is True or _check(isamAppliance) != replicating:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put("Updating web runtime component replication status",
                                            "/isam/runtime_components?cluster=true",
                                            {
                                                'replicating': replicating
                                            })

    return isamAppliance.create_return_object()


def _check(isamAppliance):
    """
    Check the replicating status
    :param isamAppliance:
    :return:
    """
    ret_obj = get(isamAppliance)

    return ret_obj['data']['replicating']



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False, ignore_error=False):
    """
    Retrieving a list of firmware settings
    """
    return isamAppliance.invoke_get("Retrieving a list of firmware settings",
                                    "/firmware_settings", ignore_error=ignore_error, requires_model=requires_model)


def backup(isamAppliance, check_mode=False, force=False):
    """
    Creating a backup of the active partition
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put("Creating a backup of the active partition",
                                        "/firmware_settings/kickoff_backup", {}, requires_model=requires_model)


def swap(isamAppliance, check_mode=False, force=False):
    """
    Swapping the active partition
    """
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        ret_obj_old = get(isamAppliance)

        ret_obj = isamAppliance.invoke_put("Swapping the active partition",
                                           "/firmware_settings/kickoff_swap", {}, requires_model=requires_model)
        # Process previous query after a successful call to swap the partition
        for partition in ret_obj_old['data']:
            if partition['active'] is False:  # Get version of inactive partition (active now!)
                ver = partition['firmware_version'].split(' ')
                isamAppliance.facts['version'] = ver[-1]

        return ret_obj


def set(isamAppliance, id, comment, check_mode=False, force=False):
    """
    Updating a comment for a partition
    """

    check_value, warnings = _check_comment(isamAppliance, comment)
    if force is True or check_value is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating a comment for a partition",
                                            f"/firmware_settings/{id}",
                                            {'comment': comment}, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_comment(isamAppliance, comment):
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']

    # Loop through firmware partitions looking for active one
    for partition in ret_obj['data']:
        if partition['active'] is True:
            return (partition['comment'] == comment), warnings

    return False, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare firmware between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['install_date']
        del obj['backup_date']
        del obj['last_boot']
    for obj in ret_obj2['data']:
        del obj['install_date']
        del obj['backup_date']
        del obj['last_boot']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2,
                                                    deleted_keys=['install_date', 'backup_date', 'last_boot'])



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get(isvgAppliance, check_mode=False, force=False):
    """
    Get host records
    """
    return isvgAppliance.invoke_get("Retrieving current host records",
                                    "/host_records")


def set(isvgAppliance, hostname, ip_addr, check_mode=False, force=False):
    """
    Add host records
    """
    if force is True or _check(isvgAppliance, hostname, ip_addr) is False:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post(
                "Setting host record",
                "/host_records",
                {
                    'hostnames': [{"name": hostname}],
                    'addr': ip_addr
                })

    return isvgAppliance.create_return_object()


def delete(isvgAppliance, hostname, ip_addr, check_mode=False, force=False):
    """
    Delete a host record
    :param isvgAppliance:
    :param hostname:
    :param ip_addr:
    :param check_mode:
    :param force:
    :return:
    """
    if force is True or _check(isvgAppliance, hostname, ip_addr) is True:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_delete(
                "Removing host record",
                "/host_records/" + ip_addr,
                {
                })
    return isvgAppliance.create_return_object()


def _check(isvgAppliance, hostname, ip_addr, check_mode=False, force=False):
    """
    check whether hostname has been defined
    :param isvgAppliance:
    :param hostname:
    :param ip_addr:
    :param check_mode:
    :param force:
    :return:
    """
    ret_obj = get(isvgAppliance)

    for hosts in ret_obj['data']:
        if hosts['addr'] == ip_addr:
            for names in hosts['hostnames']:
                if names['name'] == hostname:
                    logger.info("hostname record exists, skip")
                    return True

    return False



import logging
import os.path
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)
requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve existing fixpacks.
    """
    return isamAppliance.invoke_get("Retrieving fixpacks",
                                    "/fixpacks")


def install(isamAppliance, file, check_mode=False, force=False):
    """
    Installing a fix pack
    """
    if force is True or _check(isamAppliance, file) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post_files(
                "Installing a fix pack",
                "/fixpacks",
                [{
                    'file_formfield': 'file',
                    'filename': file,
                    'mimetype': 'application/octet-stream'
                }],
                {}, requires_model=requires_model)

    return isamAppliance.create_return_object()


def rollback(isamAppliance, file, check_mode=False, force=False):
    """
    Rollback the most recently installed fixpack
    """
    if force is True or _check_rollback(isamAppliance, file) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Rollback the most recently installed fixpack",
                "/fixpacks",
                requires_modules=None,
                requires_version="9.0.3.0",
                requires_model=requires_model
            )

    return isamAppliance.create_return_object()


def _check_rollback(isamAppliance, fixpack):
    """
    Check if fixpack is already installed
    """
    ret_obj = get(isamAppliance)

    fixpack_name = _extract_fixpack_name(fixpack)

    # Reverse sort the json by 'id'
    json_data_sorted = sorted(ret_obj['data'], key=lambda k: int(k['id']), reverse=True)

    # Eliminate all rollbacks before hitting the first non-rollback fixpack
    del_fixpack = ''  # Delete succeeding fixpack to a rollback, only last fixpack can be rolled back
    for fp in json_data_sorted:
        if fp['action'] == 'Uninstalled':
            del_fixpack = fp['name']
        elif del_fixpack == fp['name'] and fp['rollback'] == 'Yes':
            del_fixpack = ''
        elif fp['name'].lower() == fixpack_name.lower():
            return True
        # The first non-rollback fixpack needs to match the name otherwise skip rollback
        else:
            return False

    return False


def _check(isamAppliance, fixpack):
    """
    Check if fixpack is already installed
    """
    ret_obj = get(isamAppliance)

    fixpack_name = _extract_fixpack_name(fixpack)

    # Reverse sort the json by 'id'
    json_data_sorted = sorted(ret_obj['data'], key=lambda k: int(k['id']), reverse=True)

    samename_fps = []
    for fp in json_data_sorted:
        if fp['name'] == fixpack_name:
            samename_fps.append(fp)

    if len(samename_fps) > 0:
        fp_sorted = sorted(samename_fps, key=lambda k: int(k['id']), reverse=True)
        if fp_sorted[0]['action'] == 'Installed':
            return True
        else:
            return False
    else:
        return False


def _extract_fixpack_name(fixpack):
    """
    Extract fixpack name from the given fixpack
    """
    import re

    # Look for the follwing string inside the fixpack file
    # FIXPACK_NAME="9021_IPv6_Routes_fix"
    for s in ibmsecurity.utilities.tools.strings(fixpack):
        match_obj = re.search(r"FIXPACK_NAME=\"(?P<fp_name>\w+)\"", s)
        if match_obj:
            logger.info(f"Fixpack name extracted from file using strings method: {match_obj.group('fp_name')}")
            return match_obj.group('fp_name')

    # Unable to extract fixpack name from binary
    # Return fixpack name derived from the filename
    file_name = os.path.basename(fixpack)
    if file_name.find(".fixpack"):
        fixpack_name = file_name.partition(".fixpack")[0]
    else:
        fixpack_name, ext_name = file_name.split('.')
    logger.info(
        f"Fixpack name could not be extracted from binary, use name derived from the filename: {fixpack_name}")
    return fixpack_name


def compare(isamAppliance1, isamAppliance2):
    """
    Compare fixpacks between two appliances
    Sort in reverse order and remove fixpacks that were rolled back before compare
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    json_data1_sorted = sorted(ret_obj1['data'], key=lambda k: int(k['id']), reverse=True)
    del_fixpack = ''  # Delete succeeding fixpack to a rollback, only last fixpack can be rolled back
    for fixpack in json_data1_sorted:
        if fixpack['action'] == 'Uninstalled':
            del_fixpack = fixpack['name']
            del fixpack
        elif del_fixpack == fixpack['name'] and fixpack['rollback'] == 'Yes':
            del_fixpack = ''
            del fixpack
        else:
            del fixpack['date']

    json_data2_sorted = sorted(ret_obj2['data'], key=lambda k: int(k['id']), reverse=True)
    del_fixpack = ''  # Delete succeeding fixpack to a rollback, only last fixpack can be rolled back
    for fixpack in json_data2_sorted:
        if fixpack['action'] == 'Uninstalled':
            del_fixpack = fixpack['name']
            del fixpack
        elif del_fixpack == fixpack['name'] and fixpack['rollback'] == 'Yes':
            del_fixpack = ''
            del fixpack
        else:
            del fixpack['date']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging

logger = logging.getLogger(__name__)

requires_model = "Appliance"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the general configuration
    """
    return isamAppliance.invoke_get("Retrieving the general configuration", "/net/general", requires_model=requires_model)


def set(isamAppliance, hostname, check_mode=False, force=False):
    """
    Updating the general configuration
    """
    check_obj, warnings = _check(isamAppliance, hostname)

    if force is True or check_obj is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put("Updating the general configuration", "/net/general",
                                            {
                                                'hostName': hostname
                                            }, requires_model=requires_model)

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, hostname):
    ret_obj = get(isamAppliance)
    warnings = ret_obj['warnings']
    if 'hostName' in ret_obj['data']:
        if ret_obj['data']['hostName'] == hostname:
            return True, warnings
        else:
            return False, warnings
    else:
        return True, warnings



import logging
import ibmsecurity.utilities.tools
import ibmsecurity.isam.base.network.interfaces

logger = logging.getLogger(__name__)


def add(isamAppliance, label, address, maskOrPrefix, overrideSubnetChecking=False, vlanId=None, allowManagement=False,
        enabled=True, check_mode=False, force=False):
    """
    Adding an IPv4 address to an interface
    """
    add_needed = True
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            del ret_obj['objType']
            del ret_obj['type']
            for addr in ret_obj['ipv4']['addresses']:
                if addr['address'] == address:
                    add_needed = False
        else:
            warnings.append(f"Interface {label} not found, Add is not supported.")
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if add_needed is True:
        ret_obj['ipv4']['overrideSubnetChecking'] = overrideSubnetChecking
    if force is True or add_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            addr = {
                '_isNew': True,
                'objType': 'ipv4Address',
                'uuid': None,
                'address': address,
                'maskOrPrefix': maskOrPrefix,
                'allowManagement': allowManagement,
                'enabledAddress': True,
                'enabled': enabled
            }
            ret_obj['ipv4']['addresses'].append(addr)
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()


def delete(isamAppliance, label, address, vlanId=None, overrideSubnetChecking=False, check_mode=False, force=False):
    """
    Deleting an IPv4 address from an interface
    """
    delete_needed = False
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            for addr in ret_obj['ipv4']['addresses']:
                if addr['address'] == address:
                    delete_needed = True
                    ret_obj['ipv4']['addresses'].remove(addr)
                    ret_obj['ipv4']['overrideSubnetChecking'] = overrideSubnetChecking
                    break
        else:
            warnings.append(f"Interface {label} not found, Delete is not supported.")
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if force is True or delete_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()


def update(isamAppliance, label, address, new_address, maskOrPrefix, vlanId=None, allowManagement=False, enabled=True,
           overrideSubnetChecking=False, check_mode=False, force=False):
    """
    Updating an IPv4 address on an interface
    """
    update_needed = False
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            for addr in ret_obj['ipv4']['addresses']:
                if addr['address'] == address:
                    upd_addr = {
                        'uuid': addr['uuid'],
                        'address': new_address,
                        'maskOrPrefix': maskOrPrefix,
                        'allowManagement': allowManagement,
                        'enabled': enabled
                    }
                    update_needed = not (
                            ibmsecurity.utilities.tools.json_sort(addr) == ibmsecurity.utilities.tools.json_sort(
                        upd_addr))
                    if (update_needed is True):
                        ret_obj['ipv4']['addresses'].remove(addr)
                        ret_obj['ipv4']['addresses'].append(upd_addr)
                        ret_obj['ipv4']['overrideSubnetChecking'] = overrideSubnetChecking
                    break
        else:
            warnings.append(f"Interface {label} not found, Update is not supported.")
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if force is True or update_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()


def set_dhcp(isamAppliance, label, vlanId=None, enabled=False, allowManagement=False, providesDefaultRoute=False,
             routeMetric=None, check_mode=False, force=False):
    """
    Updating IPv4 dhcp on an interface
    """
    update_needed = False
    ret_obj = {}
    warnings = []
    if force is False:
        ret_obj, warnings = ibmsecurity.isam.base.network.interfaces._get_interface(isamAppliance, label, vlanId)
        if ret_obj is not None:
            upd_dhcp = {
                'enabled': enabled,
                'allowManagement': allowManagement,
                'providesDefaultRoute': providesDefaultRoute,
                'routeMetric': routeMetric
            }
            update_needed = not (
                    ibmsecurity.utilities.tools.json_sort(
                        ret_obj['ipv4']['dhcp']) == ibmsecurity.utilities.tools.json_sort(
                upd_dhcp))
            if (update_needed is True):
                ret_obj['ipv4']['dhcp'] = upd_dhcp
        else:
            warnings.append(f"Interface {label} not found, Set dhcp is not supported.")
            return isamAppliance.create_return_object(changed=False, warnings=warnings)

    if force is True or update_needed is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return ibmsecurity.isam.base.network.interfaces._update_interface(isamAppliance, ret_obj)

    return isamAppliance.create_return_object()


def search(isamAppliance, address, check_mode=False, force=False):
    """
    Retrieving uuid for an ipv4 address
    """
    ret_obj = ibmsecurity.isam.base.network.interfaces.get_all(isamAppliance)
    ipv4_address_uuid = None

    for intfc in ret_obj['data']['interfaces']:
        for adds in intfc['ipv4']['addresses']:
            if adds['address'] == address:
                ipv4_address_uuid = adds['uuid']
                break

    ret_obj = isamAppliance.create_return_object()
    ret_obj['data'] = ipv4_address_uuid

    return ret_obj



import logging
import ibmsecurity.isam.aac.server_connections.ws
from ibmsecurity.utilities.tools import json_sort
import json

logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"



def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the current general SCIM configuration settings
    """
    return isamAppliance.invoke_get("Retrieving the current general SCIM configuration settings",
                                    f"{uri}/general",
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )

def set(isamAppliance, enable_header_authentication=None, enable_authz_filter=None, admin_group=None, max_user_responses=None, attribute_modes=None, check_mode=False, force=False):
    """
    Updating the general SCIM configuration settings
    """
    ret_obj = get(isamAppliance)
    ret_obj = ret_obj['data']


    new_obj = {}


    if enable_header_authentication is None:
        new_obj['enable_header_authentication'] = ret_obj['enable_header_authentication']
    else:
        new_obj['enable_header_authentication'] = enable_header_authentication

    if enable_authz_filter is None:
        new_obj['enable_authz_filter'] = ret_obj['enable_authz_filter']
    else:
        new_obj['enable_authz_filter'] = enable_authz_filter

    if admin_group is None:
        new_obj['admin_group'] = ret_obj['admin_group']
    else:
        new_obj['admin_group'] = admin_group

    if max_user_responses is not None:
        new_obj['max_user_responses'] = max_user_responses
    else:
        if 'max_user_responses' in ret_obj:
            new_obj['max_user_responses'] = ret_obj['max_user_responses']

    if attribute_modes is not None:
        new_obj['attribute_modes'] = attribute_modes
    else:
        if 'attribute_modes' in ret_obj:
            new_obj['attribute_modes'] = ret_obj['attribute_modes']

    obj1 = json_sort(ret_obj)
    obj2 = json_sort(new_obj)


    if obj1 == obj2:
        update_required = False
    else:
        update_required = True


    if force is True or update_required is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Updating the ISAM user SCIM configuration settings",
                f"{uri}/general",
                new_obj, requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(changed=False)



import logging
import json
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/fido2/u2f-migration"
requires_modules = ["mga"]
requires_version = "9.0.7.0"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a count of U2F Registrations yet to be migrated
    """
    return isamAppliance.invoke_get("Retrieve a count of U2F Registrations yet to be migrated", uri,
                                    requires_modules=requires_modules, requires_version=requires_version)

def migrate(isamAppliance, batchSize, batchCount, check_mode=False, force=False):
    """
    Migrate U2F Registrations
    """

    json_data = {
        "batchSize": batchSize,
        "batchCount": batchCount
    }

    ret_obj = isamAppliance.invoke_post("Migrate U2F Registrations", uri, json_data,
                                    requires_modules=requires_modules, requires_version=requires_version)

    return ret_obj



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/isam/tracing"
requires_modules = None
requires_version = "10.0.8.0"
warnings=[]

def get(isamAppliance, full_list=False, deployed_policy=False, check_mode=False, force=False):
    """
    Get tracing settings
    """
    full_uri = uri + ibmsecurity.utilities.tools.create_query_string(full_list=str(full_list).lower(), deployed_policy=str(deployed_policy).lower())
    logger.debug(f"{full_uri}")
    return isamAppliance.invoke_get("Get tracing",
                                    full_uri,
                                    requires_modules=requires_modules,
                                    requires_version=requires_version,
                                    warnings=warnings)



import logging
import hashlib
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/sslcert_object"


def get_all(isvgAppliance, check_mode=False, force=False):
    """
    Get signer certificate entries
    """
    return isvgAppliance.invoke_get("Retrieving signer certificate entries", uri)


def search(isvgAppliance, label, check_mode=False, force=False):
    """
    Search for signer certificate by label
    """
    ret_obj = get_all(isvgAppliance)
    return_obj = isvgAppliance.create_return_object()

    for obj in ret_obj['data']:
        logger.debug("obj {0}".format(obj))
        if 'certAlias' in obj and obj['certAlias'] == label:
            logger.info("Found certificate entry: {0}".format(obj['certAlias']))
            return_obj['data'] = obj
            return_obj['rc'] = 0
            break

    return return_obj


def upload(isvgAppliance, certificate, label, check_mode=False, force=False):
    """
    Import signer certificate
    """
    warnings = []

    if force is True or not _check(isvgAppliance, certificate, label):
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            file_short = tools.path_leaf(certificate);
            ret_obj = isvgAppliance.invoke_post_files(
                "Import signer certificate",
                "/upload_object/" + file_short,
                [
                    {
                        'file_formfield': 'uploadedfile',
                        'filename': certificate,
                        'mimetype': 'application/x-x509-ca-cert'
                    }
                ],
                {
                    'certAlias': label,
                    'certLocation': file_short,
                    'upload_file_type': 'ssl_cert'
                },
                json_response=False)

            warnings = ret_obj['warnings']

            if ret_obj["rc"] == 0:
                certificate_json = {
                   "name": label,
                   "certLocation": file_short,
                   "certAlias": label,
                   "_isNew": True,
                   "action":"configure",
                   "uploadedfile":[
                      {
                         "index":0,
                         "name": file_short,
                         "type": "application/x-x509-ca-cert"
                      }
                   ]
                }

                return isvgAppliance.invoke_post(
                    "Save signer certificate metadata", uri + "/", certificate_json, warnings=warnings)

    return isvgAppliance.create_return_object()


def delete(isvgAppliance, certificate, check_mode=False, force=False):
    """
    Delete signer certificate
    """
    warnings = []

    ret_obj = search(isvgAppliance, label, check_mode=check_mode, force=force)

    if force is True or ret_obj['data'] != {}:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            if 'uuid' in ret_obj['data']:
                uuid = ret_obj['data']['uuid']
                warnings = ret_obj['warnings']
                return isvgAppliance.invoke_delete(
                    "Delete an external certificate", "{0}/{1}".format(uri, uuid), warnings=warnings)

    return isvgAppliance.create_return_object(warnings=warnings)


def _check(isvgAppliance, certificate, label):
    """
    Check if signer certificate needs to be uploaded
    """
    ret_obj = search(isvgAppliance, label)

    if len(ret_obj['data']) > 0 and 'checkSum' in ret_obj['data']:
        applianceCheckSum = ret_obj['data']['checkSum']
        with open(certificate, "rb") as f:
            bytes = f.read()
        localCheckSum = hashlib.sha256(bytes).hexdigest()
        logger.debug("applianceCheckSum={0}".format(applianceCheckSum))
        logger.debug("localCheckSum={0}".format(localCheckSum))
        if applianceCheckSum == localCheckSum:
            return True
        else:
            return False
    else:
        return False



import logging
import os
from io import open
import ibmsecurity.isam.web.reverse_proxy.instance as rp

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/iag/export"
requires_modules = ["wga"]
requires_version = "10.0.4"

def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the wga features for iag
    """
    return isamAppliance.invoke_get("Retrieving the IAG features", f"{uri}/features",
                                    requires_version=requires_version, requires_modules=requires_modules)

def get(isamAppliance, instance, check_mode=False, force=False):
    """
    Retrieving the wga junctions for a specific instance for iag
    """
    if rp._check(isamAppliance, instance):
        return isamAppliance.invoke_get(f"Retrieving the IAG junction for {instance}", f"{uri}/{instance}/junctions",
                                        requires_version=requires_version, requires_modules=requires_modules)
    else:
        logger.debug(f"Instance: {instance} does not exist")
        return isamAppliance.create_return_object(warnings=f"Instance: {instance} does not exist", rc=1)


def download(isamAppliance, instance, junctions, features, filename="iag.zip", check_mode=False, force=False):
    """
    Downloads the configuration for the instance and junction and features
    It will delete the file if it exists already only if force=True.
    """
    if force or rp._check(isamAppliance, instance):
        if force and os.path.exists(filename):
          os.remove(filename)
          logger.debug(f"Removed {filename}")

        if not check_mode:
          json_data = {
               "junctions": junctions,
               "features": features
               }
          zip_data = isamAppliance.invoke_post(f"Exporting junctions and features for IAG for {instance}", f"{uri}/{instance}/download", data=json_data,
                                               requires_version = requires_version, requires_modules = requires_modules)
          with open(filename, 'wb') as f:
             f.write(zip_data['data'])
        else:
          return isamAppliance.create_return_object(warnings=f"Skipped export, we are in check_mode {check_mode}",
                                                      rc=0)
    else:
        logger.debug(f"Instance: {instance} does not exist, cannot export")
        return isamAppliance.create_return_object(warnings=f"Instance: {instance} does not exist, cannot export", rc=1)

def validate(isamAppliance, instance, junctions, features, check_mode=False, force=False):
    """
    Returns an overview of the junctions and features with warnings.
    """
    if rp._check(isamAppliance, instance):
        json_data = {
            "junctions": junctions,
            "features": features
        }

        return isamAppliance.invoke_post(f"Exporting junctions and features for IAG for {instance}", f"{uri}/{instance}/validate", data=json_data,
                                         requires_version=requires_version, requires_modules=requires_modules)

    else:
        logger.debug(f"Instance: {instance} does not exist, cannot export")
        return isamAppliance.create_return_object(warnings=f"Instance: {instance} does not exist, cannot export", rc=1)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/cert_object"


def get(isvgAppliance, check_mode=False, force=False):
    """
    Get personal cert store entry
    """
    return isvgAppliance.invoke_get("Retrieving cert store entry", uri)


def search(isvgAppliance, serial, check_mode=False, force=False):
    """
    Search for existing certificate by serial
    """
    ret_obj = get(isvgAppliance)
    return_obj = isvgAppliance.create_return_object()

    if 'serialnumber' in ret_obj['data'] and ret_obj['data']['serialnumber'] == int(serial):
        logger.info("Found cert entry: {0}".format(ret_obj['data']['serialnumber']))
        return_obj['data'] = ret_obj['data']
        return_obj['rc'] = 0

    return return_obj


def upload(isvgAppliance, db, type, password, serial, check_mode=False, force=False):
    """
    Import personal certificate from db
    """
    warnings = []

    file_short = tools.path_leaf(db);
    if force is True or not _check(isvgAppliance, serial):
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            ret_obj = isvgAppliance.invoke_post_files(
                "Import personal certificate",
                "/upload_object",
                [
                    {
                        'file_formfield': 'uploadedfile',
                        'filename': db,
                        'mimetype': 'application/x-pkcs12'
                    }
                ],
                {
                    'keyFilePassword': password,
                    'upload_file_type': 'app_ssl_cert'
                },
                json_response=False)

            warnings = ret_obj['warnings']

            if ret_obj["rc"] == 0:
                cert_json = {
                   "fileName": file_short,
                   "keyFilePassword": password,
                   "storeType": type
                }

                return isvgAppliance.invoke_post(
                    "Save personal certificate", "/cert_object/", cert_json, warnings=warnings)

    return isvgAppliance.create_return_object()


def _check(isvgAppliance, serial):
    """
    Check if personal cert needs to be uploaded
    """
    ret_obj = search(isvgAppliance, serial)

    if len(ret_obj['data']) > 0:
        return True
    else:
        return False



import logging
import ibmsecurity.utilities.tools as _tools
import json

logger = logging.getLogger(__name__)


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the administrator settings
    """
    return isamAppliance.invoke_get("Retrieving the administrator settings", "/admin_cfg")


def set_pw(isamAppliance, oldPassword, newPassword, sessionTimeout="30", httpsPort=None, check_mode=False, force=False):
    """
    Set password for admin user (super user for appliance)

    Note: this function is being maintained for backward compatibility. Use set() going forward.
    """
    warnings = ["Password change requested - cannot query existing password for idempotency check."]
    if check_mode is True:
        return isamAppliance.create_return_object(changed=True, warnings=warnings)
    else:
        json_data = {
            "oldPassword": oldPassword,
            "newPassword": newPassword,
            "confirmPassword": newPassword,
            "sessionTimeout": sessionTimeout
        }
        if httpsPort is not None:
            if _tools.version_compare(isamAppliance.facts['version'], "9.0.1.0") < 0:
                warnings.append(
                    "Appliance at version: {0}, httpsPort not supported. Needs 9.0.1.0 or higher. Ignoring httpsPort for this call.")
            else:
                json_data['httpsPort'] = httpsPort
        else:
            if _tools.version_compare(isamAppliance.facts['version'], "9.0.1.0") < 0:
                pass  # Can safely ignore httpsPort
            else:
                warnings.append("Default httpsPort of 443 will be set on the appliance.")
        return isamAppliance.invoke_put("Setting admin password", "/core/admin_cfg", json_data, warnings=warnings)


def set(isamAppliance,
        oldPassword=None,
        newPassword=None,
        minHeapSize=None,
        maxHeapSize=None,
        sessionTimeout=30,
        sessionInactivityTimeout=None,
        sessionCachePurge=None,
        baSessionTimeout=None,
        httpPort=None,
        httpsPort=None,
        sshdPort=None,
        sshdClientAliveInterval=None,
        sshdPasswordAuthentication=None,
        swapFileSize=None,
        minThreads=None,
        maxThreads=None,
        maxPoolSize=None,
        lmiDebuggingEnabled=None,
        consoleLogLevel=None,
        acceptClientCerts=None,
        validateClientCertIdentity=None,
        excludeCsrfChecking=None,
        enabledServerProtocols=None,
        enabledTLS=None,
        enableSSLv3=None,
        maxFiles=None,
        maxFileSize=None,
        httpProxy=None,
        httpsProxy=None,
        loginHeader=None,
        loginMessage=None,
        pendingChangesLifetime=None,
        accessLogFormat=None,
        lmiMessageTimeout=None,
        validVerifyDomains=None,
        jsVersion=None,
        check_mode=False, force=False):
    """
    Updating the administrator settings
    """
    warnings = []

    update_required, warnings, json_data = _check(isamAppliance,
                                                  warnings=warnings,
                                                  oldPassword=oldPassword,
                                                  newPassword=newPassword,
                                                  minHeapSize=minHeapSize,
                                                  maxHeapSize=maxHeapSize,
                                                  sessionTimeout=sessionTimeout,
                                                  sessionInactivityTimeout=sessionInactivityTimeout,
                                                  sessionCachePurge=sessionCachePurge,
                                                  baSessionTimeout=baSessionTimeout,
                                                  httpPort=httpPort,
                                                  httpsPort=httpsPort,
                                                  sshdPort=sshdPort,
                                                  sshdClientAliveInterval=sshdClientAliveInterval,
                                                  sshdPasswordAuthentication=sshdPasswordAuthentication,
                                                  swapFileSize=swapFileSize,
                                                  minThreads=minThreads,
                                                  maxThreads=maxThreads,
                                                  maxPoolSize=maxPoolSize,
                                                  lmiDebuggingEnabled=lmiDebuggingEnabled,
                                                  consoleLogLevel=consoleLogLevel,
                                                  acceptClientCerts=acceptClientCerts,
                                                  validateClientCertIdentity=validateClientCertIdentity,
                                                  excludeCsrfChecking=excludeCsrfChecking,
                                                  enabledServerProtocols=enabledServerProtocols,
                                                  enabledTLS=enabledTLS,
                                                  enableSSLv3=enableSSLv3,
                                                  maxFiles=maxFiles,
                                                  maxFileSize=maxFileSize,
                                                  httpProxy=httpProxy,
                                                  httpsProxy=httpsProxy,
                                                  loginHeader=loginHeader,
                                                  loginMessage=loginMessage,
                                                  pendingChangesLifetime=pendingChangesLifetime,
                                                  accessLogFormat=accessLogFormat,
                                                  lmiMessageTimeout=lmiMessageTimeout,
                                                  validVerifyDomains=validVerifyDomains,
                                                  jsVersion=jsVersion
                                                  )

    if force or update_required:
        if check_mode:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Updating the administrator settings",
                "/core/admin_cfg", json_data, warnings=warnings)

    return isamAppliance.create_return_object(changed=False, warnings=warnings)


def _check(isamAppliance,
           warnings=[],
           **kwargs
           ):
    """
    Check whether target key has already been set with the value
    :param isamAppliance:
    :param key:
    :param value:
    :return: True/False
    """
    ret_obj = get(isamAppliance)

    json_data = {}
    for k, v in kwargs.items():
        iviaVersion = isamAppliance.facts['version']
        if v is None:
            # Skip None inputs
            continue
        if k == "force":
            # force = v
            continue
        if k == "check_mode":
            # check_mode = v
            continue
        if k not in ret_obj['data']:
            # add as empty string, if it's not there. I know this is true for jsVersion, and probably also for other stuff
            ret_obj['data'][k] = ""
        if k == "newPassword":
            json_data["confirmPassword"] = v
        if k in ["minHeapSize", "maxHeapSize", "httpPort", "httpsPort", "minThreads", "maxThreads", "maxPoolSize", "maxFiles", "maxFileSize", "sshdPort", "sessionCachePurge", "sessionInactivityTimeout", "sshdClientAliveInterval", "baSessionTimeout"]:
            # int values
            if k == "sshdPort" and _tools.version_compare(iviaVersion, "9.0.3.0") < 0:
                warnings.append(f"Appliance at version: {iviaVersion}, sshdPort: {v} is not supported. Needs 9.0.3.0 or higher. Ignoring sshdPort for this call.")
                continue
            if k in ["sessionCachePurge", "sessionInactivityTimeout", "sshdClientAliveInterval"] and _tools.version_compare(iviaVersion, "9.0.5.0") < 0:
                warnings.append(f"Appliance at version: {iviaVersion}, {k}: {v} is not supported. Needs 9.0.5.0 or higher. Ignoring.")
                continue
            if k in ["baSessionTimeout"] and _tools.version_compare(iviaVersion, "10.0.2.0") < 0:
                warnings.append(
                    f"Appliance at version: {iviaVersion}, {k}: {v} is not supported. Needs 10.0.2.0 or higher. Ignoring.")
                continue
            json_data[k] = int(v)
            continue
        if k == "enableSSLv3":
            if _tools.version_compare(iviaVersion, "10.0.3.0") >= 0:
                warnings.append(f"Appliance at version: {iviaVersion}, enableSSLv3: {v} is not supported. Needs max. 10.0.2.0. Ignoring for this call.")
                continue
        if k == "consoleLogLevel":
            if 'consoleLogLevel' in ret_obj['data'] and ret_obj['data']['consoleLogLevel'] == 'OFF':
                ret_obj['data']['consoleLogLevel'] = 'OFF'
        if k == "enabledTLS":
            if _tools.version_compare(iviaVersion, "9.0.4.0") < 0:
                warnings.append(f"Appliance at version: {iviaVersion}, enabledTLS: {v} is not supported. Needs 9.0.4.0 or higher. Ignoring enabledTLS for this call.")
                continue
        if k in ["swapFileSize", "httpProxy"]:
            if _tools.version_compare(iviaVersion, "9.0.5.0") < 0:
                warnings.append(f"Appliance at version: {iviaVersion}, {k}: {v} is not supported. Needs 9.0.5.0 or higher. Ignoring.")
                continue
        if k in ["enabledServerProtocols", "loginHeader", "loginMessage", "pendingChangesLifetime", "httpsProxy"]:
            if _tools.version_compare(iviaVersion, "9.0.7.0") < 0:
                warnings.append(
                    f"Appliance at version: {iviaVersion}, {k}: {v} is not supported. Needs 9.0.7.0 or higher. Ignoring.")
                continue
        if k in ["accessLogFormat"]:
            if _tools.version_compare(iviaVersion, "10.0.0.0") < 0:
                warnings.append(f"Appliance at version: {iviaVersion}, {k}: {v} is not supported. Needs 10.0.0.0 or higher. Ignoring.")
                continue
        if k in ["lmiMessageTimeout", "validVerifyDomains"]:
            if _tools.version_compare(iviaVersion, "10.0.2.0") < 0:
                warnings.append(f"Appliance at version: {iviaVersion}, {k}: {v} is not supported. Needs 10.0.0.0 or higher. Ignoring.")
                continue
        if k == "jsVersion":
            if _tools.version_compare(iviaVersion, "10.0.9.0") < 0:
                warnings.append(f"Appliance at version: {iviaVersion}, {k}: {v} is not supported. Needs 10.0.9.0 or higher. Ignoring.")
                continue

        # Add to the json_data dict
        json_data[k] = v

    if not _tools.json_equals(ret_obj, json_data):
        logger.debug("Admin Settings are found to be different. See above JSON for difference.")
        return True, warnings, json_data
    else:  # No changes required
        return False, warnings, json_data


def compare(isamAppliance1, isamAppliance2):
    """
    Compare advanced tuning parameters between two appliances
    """
    ret_obj1 = get(isamAppliance1)
    ret_obj2 = get(isamAppliance2)

    return _tools.json_compare(ret_obj1, ret_obj2)



import logging
from ibmsecurity.appliance.ibmappliance import IBMError
from ibmsecurity.utilities.tools import json_compare

logger = logging.getLogger(__name__)

requires_model = "Appliance"

try:
    basestring
except NameError:
    basestring = (str, bytes)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the list of host IP addresses
    """
    return isamAppliance.invoke_get("Retrieving the list of host IP addresses",
                                    "/isam/host_records", requires_model=requires_model)


def get(isamAppliance, addr, check_mode=False, force=False):
    """
    Retrieving the list of host names associated with a host IP address
    """
    return isamAppliance.invoke_get(f"Retrieving the list of host names associated with a host IP address {addr}",
                                    f"/isam/host_records/{addr}/hostnames", requires_model=requires_model)


def set(isamAppliance, addr, hostnames, check_mode=False, force=False):
    """
    Creates or updates specified address with supplied hostnames.

    :param isamAppliance:
    :param addr:          String, addr of host record to update
    :param hostnames:     Array, hostnames to be added
    :param check_mode:
    :param force:
    :return:
    """

    # Make sure if only one hostname is entered that it is an array
    if isinstance(hostnames, basestring):
        hostnames = [hostnames]

    ret_obj = isamAppliance.create_return_object()
    exists = False
    hostnames_remaining = hostnames
    addr_exists = True

    if force is False:
        exists, hostnames_remaining, addr_exists, warnings = _check(isamAppliance, addr, hostnames)
        ret_obj['warnings'] = warnings

    if force or exists is False:
        if check_mode is True:
            ret_obj['changed'] = True
        else:
            if addr_exists:
                # Call update for each hostname
                for name in hostnames_remaining:
                    # Forcing because we've already checked that the hostnames don't exist
                    # TODO: determine how best to return this, since it makes multiple requests
                    update(isamAppliance, addr, name, check_mode, force=True)
                    ret_obj['changed'] = True
            else:
                # Forcing because we've already checked that the hostnames don't exist
                return add(isamAppliance, addr, hostnames_remaining, check_mode, force=True)

    return ret_obj


def add(isamAppliance, addr, hostnames, check_mode=False, force=False):
    """
    Creating a host record (IP address and host name)
    :param isamAppliance:
    :param addr:          String, addr of host record to update
    :param hostnames:     Array, hostnames to be added
    :param check_mode:
    :param force:
    :return:
    """

    # Make sure if only one hostname is entered that it is an array
    if isinstance(hostnames, basestring):
        hostnames = [hostnames]

    exists = False
    hostnames_remaining = hostnames
    addr_exists = True

    if force is False:
        exists, hostnames_remaining, addr_exists, warnings = _check(isamAppliance, addr, hostnames)

    if force or addr_exists is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            # Format POST response correctly
            hostnames_post = []
            for entry in hostnames_remaining:
                hostnames_post.append({'name': entry})

            return isamAppliance.invoke_post(
                "Creating a host record (IP address and host name)",
                "/isam/host_records",
                {
                    'addr': addr,
                    'hostnames': hostnames_post
                }, requires_model=requires_model)
    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, addr, name, check_mode=False, force=False):
    """
    Adding a host name to a host IP address

    :param isamAppliance:
    :param addr:          String, addr of host record to update
    :param name:          String, hostname to add to addr
    :return:
    """

    exists = False
    hostnames_remaining = [name]
    addr_exists = True

    if force is False:
        exists, hostnames_remaining, addr_exists, warnings = _check(isamAppliance, addr, [name])
        if addr_exists is False:
            if warnings == []:
                raise IBMError("HTTP Return code: 404",
                               "Specified addr does not exist in the hosts file, cannot update.")
            else:
                return isamAppliance.create_return_object(warnings=warnings)

    if force or exists is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_post(
                "Adding a host name to a host IP address",
                f"/isam/host_records/{addr}/hostnames",
                {'name': hostnames_remaining[0]},
                requires_model=requires_model
            )

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, addr, name=None, check_mode=False, force=False):
    """
    Removing a host name from a host IP address
    :param isamAppliance:
    :param addr:          String, addr of host record to update
    :param name:          String, hostname to add to addr
    :param check_mode:
    :param force:
    :return:
    """

    exists = False
    addr_exists = True

    if force is False:
        exists, _, addr_exists, warnings = _check(isamAppliance, addr, [name])
        if addr_exists is False:
            return isamAppliance.create_return_object(warnings=warnings)

    if force or exists is True or name is None:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            if exists is True:
                # Delete specific entry
                return isamAppliance.invoke_delete(
                    "Removing a host name from a host IP address",
                    f"/isam/host_records/{addr}/hostnames/{name}",
                    requires_model=requires_model
                )
            elif exists is False and name is None:
                # Delete all entries for address if no name given
                return isamAppliance.invoke_delete(
                    "Removing a host record (IP address and associated host names)",
                    f"/isam/host_records/{addr}",
                    requires_model=requires_model
                )

    return isamAppliance.create_return_object(warnings=warnings)


def _check(isamAppliance, addr, hostnames=[]):
    """
    Check addr has all hostnames set
    """
    exists = False
    addr_exists = False

    # Make sure addr exists in host records

    ret_obj = get_all(isamAppliance)
    warnings = ret_obj['warnings']

    for obj in ret_obj['data']:
        if obj['addr'] == addr:
            addr_exists = True
            if len(hostnames) > 0:
                for entry in obj['hostnames']:
                    if entry['name'] in hostnames:
                        hostnames.remove(entry['name'])
                        logger.debug("Host record exists, continuing checking")

    if addr_exists is False:
        logger.debug("Host record does not exist")
        return exists, hostnames, addr_exists, warnings
    else:
        if len(hostnames) < 1:
            exists = True
            logger.info("All host records exist")
        else:
            logger.info("Not all host records exist")

    return exists, hostnames, addr_exists, warnings


def compare(isamAppliance1, isamAppliance2):
    """
    Compare host records between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    # # Ignore differences between comments/uuid as they are immaterial
    # for param in ret_obj1['data']['tuningParameters']:
    #     del param['comment']
    #     del param['uuid']
    # for param in ret_obj2['data']['tuningParameters']:
    #     del param['comment']
    #     del param['uuid']

    return json_compare(ret_obj1, ret_obj2)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)

uri = "/setup_complete_guided"

def get(isvgAppliance, check_mode=False, force=False):
    """
    Get Set up complete status
    """
    return isvgAppliance.invoke_get("Get Setup Complete Settings", "/")


def complete(isvgAppliance, check_mode=False, force=False):
    """
    Setup complete
    """
    if force is False:
        ret_obj = get(isvgAppliance)

    #
    # if the guided setup is complete, fetching data at '/' will return non-empty JSON payload
    if force is True or len(ret_obj['data']) == 0:
        if check_mode is True:
            return isvgAppliance.create_return_object(changed=True)
        else:
            return isvgAppliance.invoke_post(
                "Setup complete",
                uri,
                {})

    return isvgAppliance.create_return_object()



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/wga/kerberos/config/realms"
requires_modules = ['wga']
requires_version = None
import ibmsecurity.isam.web.kerberos_configuration.realms as realms


def search(isamAppliance, realm, subsection, check_mode=False, force=False):
    """
    Search kerberos realm subsection by name
    """
    logger.info("subsection to search under: {0} is: {1} ".format(realm, subsection))
    ret_obj = realms.get(isamAppliance, realm)
    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['type'] == "section":
            if obj['name'] == subsection:
                logger.info("Found Kerberos realm subsection {0} id: {1}".format(subsection, obj['id']))
                return_obj['data'] = obj
                return_obj['rc'] = 0
                break

    return return_obj


def _check(isamAppliance, realm, subsection):
    """
    Check if kerberos realm's subsection is already created or not

    :param isamAppliance:
    :return:
    """
    ret_obj = search(isamAppliance, realm, subsection)

    realm = "{0}/{1}".format("realms", realm)

    logger.debug("Looking for existing kerberos subsection: {1} in realm: {0} in: {2}".format(realm, subsection,
                                                                                              ret_obj['data']))
    if ret_obj['data'] != {}:
        if ret_obj['data']['name'] == subsection:
            logger.debug("Found kerberos realm's subsection: {0}".format(subsection))
            return True
    return False


def add(isamAppliance, realm, subsection, check_mode=False, force=False):
    """
        Add kerberos realm's subsection
        :param isamAppliance:
        :return:
     """
    check_realm = realms.search(isamAppliance, realm)

    if check_realm == {}:
        return isamAppliance.create_return_object(warnings=["Kerberos realm: {0} not found".format(realm)])

    check_subsection = _check(isamAppliance, realm, subsection)

    if check_subsection is True and force is False:
        return isamAppliance.create_return_object(
            warnings=["Kerberos subsection {1} in realm: {0} already exists".format(realm, subsection)])

    if force is True or check_subsection is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post("Add kerberos realm's subsection", "{0}/{1}".format(uri, realm),
                                             {
                                                 "subsection": subsection
                                             },
                                             requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, realm, subsection, check_mode=False, force=False):
    """
    Remove kerberos realms subsection
    :param isamAppliance:
    :return:
    """
    logger.debug(" Remove param uri = {0}/{1}/{2}".format(uri, realm, subsection))

    if realms.search(isamAppliance, realm) == {} and force is False:
        return isamAppliance.create_return_object(warnings=["Realm: {0} does not exists: ".format(realm)])

    if _check(isamAppliance, realm, subsection) is False and force is False:
        return isamAppliance.create_return_object(warnings=["subsection: {0} not found".format(subsection)])
    else:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(description="Delete kerberos realm subsection ",
                                               uri="{0}/{1}/{2}".format(uri, realm, subsection),
                                               requires_modules=requires_modules, requires_version=requires_version)



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/mga/server_connections/ws"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2.1"  # Will change if introduced in an earlier version.


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Retrieving a list of all Web Service connections
    """
    return isamAppliance.invoke_get("Retrieving a list of all Web Service connections",
                                    f"{uri}/v1", requires_modules=requires_modules,
                                    requires_version=requires_version)


def get(isamAppliance, name=None, check_mode=False, force=False):
    """
    Retrieving a Web Service connection
    """
    ret_obj = search(isamAppliance, name=name, force=force)
    id = ret_obj["data"]

    if id == {}:
        logger.info(f"Web Service connection {name} had no match, skipping retrieval.")
        return isamAppliance.create_return_object()
    else:
        return isamAppliance.invoke_get("Retrieving a Web Service connection",
                                        f"{uri}/{id}/v1",
                                        requires_modules=requires_modules,
                                        requires_version=requires_version)


def search(isamAppliance, name, check_mode=False, force=False):
    """
    Search UUID for named Web Service connection
    """
    ret_obj = get_all(isamAppliance)
    return_obj = isamAppliance.create_return_object()
    return_obj["warnings"] = ret_obj["warnings"]

    for obj in ret_obj['data']:
        if obj['name'] == name:
            logger.info(f"Found Web Service connection {name} id: {obj['uuid']}")
            return_obj['data'] = obj['uuid']
            return_obj['rc'] = 0

    return return_obj


def add(isamAppliance, name, connection, description='', locked=False, check_mode=False, force=False):
    """
    Creating a Web Service connection
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Creating a Web Service connection",
                f"{uri}/v1",
                _create_json(name=name, description=description, locked=locked, connection=connection),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()


def delete(isamAppliance, name, check_mode=False, force=False):
    """
    Deleting a Web Service connection
    """
    ret_obj = search(isamAppliance, name, check_mode=check_mode, force=force)
    id = ret_obj["data"]

    if force is True or _check_exists(isamAppliance, name=name) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a Web Service connection",
                f"{uri}/{id}/v1", requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object()


def update(isamAppliance, name, connection, description='', locked=False, new_name=None, ignore_password_for_idempotency=False,
           check_mode=False, force=False):
    """
    Modifying a Web Service connection

    Use new_name to rename the connection
    """
    ret_obj = get(isamAppliance, name)
    warnings = ret_obj["warnings"]

    if ret_obj["data"] == {}:
        warnings.append(f"Web Service connection {name} not found, skipping update.")
        return isamAppliance.create_return_object(warnings=warnings)
    else:
        id = ret_obj["data"]["uuid"]

    needs_update = False

    json_data = _create_json(name=name, description=description, locked=locked, connection=connection)
    if new_name is not None:  # Rename condition
        json_data['name'] = new_name

    if force is not True:
        if 'uuid' in ret_obj['data']:
            del ret_obj['data']['uuid']
        if ignore_password_for_idempotency:
            if 'password' in connection:
                warnings.append("Request made to ignore password for idempotency check.")
                connection.pop('password', None)

        sorted_ret_obj = tools.json_sort(ret_obj['data'])
        sorted_json_data = tools.json_sort(json_data)
        logger.debug(f"Sorted Existing Data:{sorted_ret_obj}")
        logger.debug(f"Sorted Desired  Data:{sorted_json_data}")
        if sorted_ret_obj != sorted_json_data:
            needs_update = True

        if 'password' in connection:
            warnings.append("Since existing password cannot be read - this call will not be idempotent.")
            needs_update = True

    if force is True or needs_update is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True, warnings=warnings)
        else:
            return isamAppliance.invoke_put(
                "Modifying a Web Service connection",
                f"{uri}/{id}/v1", json_data, requires_modules=requires_modules,
                requires_version=requires_version, warnings=warnings)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, name, connection, description='', locked=False, new_name=None, ignore_password_for_idempotency=False, check_mode=False, force=False):
    """
    Creating or Modifying a Web Service connection
    """
    if (search(isamAppliance, name=name))['data'] == {}:
        # Force the add - we already know connection does not exist
        return add(isamAppliance=isamAppliance, name=name, connection=connection, description=description,
                   locked=locked, check_mode=check_mode, force=True)
    else:
        # Update request
        return update(isamAppliance=isamAppliance, name=name, connection=connection, description=description,
                      locked=locked, new_name=new_name, ignore_password_for_idempotency=ignore_password_for_idempotency, check_mode=check_mode, force=force)


def _create_json(name, description, locked, connection):
    """
    Create a JSON to be used for the REST API call
    """

    json = {
        "connection": connection,
        "type": "ws",
        "name": name,
        "description": description,
        "locked": locked
    }

    return json


def compare(isamAppliance1, isamAppliance2):
    """
    Compare Web Service connections between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']:
        del obj['uuid']
    for obj in ret_obj2['data']:
        del obj['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])


def _check_exists(isamAppliance, name=None, id=None):
    """
    Check if WS Connection already exists
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if (name is not None and obj['name'] == name) or (id is not None and obj['uuid'] == id):
            return True

    return False



import logging

logger = logging.getLogger(__name__)
uri = "/setup_service_agreements"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the service agreement settings
    """
    return isamAppliance.invoke_get("Retrieving the service agreement settings", "/setup_service_agreements/accepted")


def set(isamAppliance, check_mode=False, force=False):
    """
    Accept Service Agreement
    """

    if force is True or _check(isamAppliance) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Accept service agreements",
                "/setup_service_agreements/accepted",
                {
                    "accepted": True
                })

    return isamAppliance.create_return_object(changed=False)


def get_non_ibm(isamAppliance, offering, check_mode=False, force=False):
    """
    Reading non-IBM the software license agreement terms
    """
    return isamAppliance.invoke_get("Reading non-IBM the software license agreement terms",
                                    f"{uri}/non_ibm_text/{offering}")


def get_terms(isamAppliance, check_mode=False, force=False):
    """
    Reading the software license agreement terms
    """
    return isamAppliance.invoke_get("Reading the software license agreement terms", f"{uri}/")


def _check(isamAppliance):
    ret_obj = get(isamAppliance)

    rc = False

    if ret_obj['data']['accepted']:
        logger.info("service agreements already accepted")
        rc = True

    return rc



import logging
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

# URI for this module
uri = "/iam/access/v8/dynamic_clients"
requires_modules = ["mga"]
requires_version = "9.0.5.0"


def get_all(isamAppliance, sortBy=None, count=None, start=None, filter=None, check_mode=False, force=False):
    """
    Retrieve a list of API protection clients
    """
    return isamAppliance.invoke_get("Retrieve several dynamically registered clients",
                                    f"{uri}/{tools.create_query_string(filter=filter, sortBy=sortBy, count=count, start=start)}",
                                    requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, client_id, check_mode=False, force=False):
    """
    Retrieve a dynamically registered client
    """
    return isamAppliance.invoke_get("Retrieve a dynamically registered client", f"{uri}/{client_id}",
                                    requires_modules=requires_modules, requires_version=requires_version)


def search(isamAppliance, client_id, check_mode=False, force=False):
    """
    Search API Protection Dynamic Client by name
    """
    ret_obj = get_all(isamAppliance)

    for obj in ret_obj['data']:
        if obj['clientId'] == client_id:
            logger.info(f"Found API Protection Dynamic Client id: {client_id}")
            return True

    return False


def delete(isamAppliance, client_id, check_mode=False, force=False):
    """
    Delete a dynamically registered client
    """

    if force or search(isamAppliance, client_id, check_mode=check_mode, force=force):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete a dynamically registered client", f"{uri}/{client_id}",
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object()



import logging


logger = logging.getLogger(__name__)

uri = "/mga/scim/configuration"
requires_modules = ["mga", "federation"]
requires_version = "9.0.2"

def get(isamAppliance, ldap_connection, check_mode=False, force=False):
    """
    Retrieving the current list of ldap object classes from the configured ldap
    """
    return isamAppliance.invoke_get("Retrieving the current list of ldap object classes from the configured ldap",
                                    "{0}/urn:ietf:params:scim:schemas:core:2.0:User/ldap_objectclasses?ldap_connection={1}".format(
                                        uri, ldap_connection),
                                    requires_modules=requires_modules,
                                    requires_version=requires_version
                                    )



from setuptools import setup, find_packages

setup(
    name="ibmsecurity",
    packages=find_packages(exclude=["test.*","test"]),
    # Date of release used for version - please be sure to use YYYY.MM.DD.seq#, MM and DD should be two digits e.g. 2017.02.05.0
    # seq# will be zero unless there are multiple release on a given day - then increment by one for additional release for that date
    version="2026.1.14.0",
    description="Idempotent functions for IBM Verify Appliance REST APIs",
    author="IBM",
    author_email="secorch@wwpdl.vnet.ibm.com",
    url="",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Build Tools",
    ],
    zip_safe=False,
    install_requires=["requests", "jmespath"],
)



import logging

from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/groups"
requires_modules = ["wga"]
requires_version = "9.0.7"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieve a list of all ISAM groups
    """
    return isamAppliance.invoke_get("Retrieve a list of all ISAM groups",
                                    "{0}".format(uri),
                                    requires_modules=requires_modules, requires_version=requires_version)



import logging
import ibmsecurity.utilities.tools

logger = logging.getLogger(__name__)


def get_all(isamAppliance, check_mode=False, force=False):
    """
    Get all email objects
    """
    return isamAppliance.invoke_get("Get all email objects",
                                    "/core/rsp_email_objs")


def get(isamAppliance, uuid, check_mode=False, force=False):
    """
    Get a specific email object
    """
    return isamAppliance.invoke_get("Get a specific email object",
                                    f"/core/rsp_email_objs/{uuid}")


def add(isamAppliance, name, smtpServer, from_email, to_email, smtpPort=25, objType='email',
        comment='', check_mode=False, force=False):
    """
    Add an email object
    """
    if force is True or _check(isamAppliance, None, name, comment, smtpServer, smtpPort, from_email, to_email) is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Add an email object",
                "/core/rsp_email_objs/",
                {
                    'name': name,
                    'objType': objType,
                    'comment': comment,
                    'smtpServer': smtpServer,
                    'smtpPort': smtpPort,
                    'from': from_email,
                    'to': to_email
                })

    return isamAppliance.create_return_object()


def update(isamAppliance, uuid, name, smtpServer, from_email, to_email, smtpPort=25, objType='email',
           comment='', check_mode=False, force=False):
    """
    Update a specific email object
    """
    if force is True or (
            _exists(isamAppliance, uuid) is True and _check(isamAppliance, uuid, name, comment, smtpServer,
                                                            smtpPort,
                                                            from_email, to_email) is False):
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_put(
                "Update a specific email object",
                f"/core/rsp_email_objs/{uuid}",
                {
                    'name': name,
                    'uuid': uuid,
                    'objType': objType,
                    'comment': comment,
                    'smtpServer': smtpServer,
                    'smtpPort': smtpPort,
                    'from': from_email,
                    'to': to_email
                })

    return isamAppliance.create_return_object()


def delete(isamAppliance, uuid, check_mode=False, force=False):
    """
    Delete an email object
    """
    if force is True or _exists(isamAppliance, uuid) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Delete an email object",
                f"/core/rsp_email_objs/{uuid}")

    return isamAppliance.create_return_object()


def _exists(isamAppliance, uuid):
    """
    Check if an uuid object exists

    :param isamAppliance:
    :param uuid:
    :return:
    """
    exists = False
    ret_obj = get_all(isamAppliance)

    for email in ret_obj['data']['emailObjects']:
        if email['uuid'] == uuid:
            exists = True
            break

    return exists


def _check(isamAppliance, uuid, name, comment, smtpServer, smtpPort, from_email, to_email):
    """
    Check if the email object exists and is the same - uuid=None means add versus delete

    NOTE: if UUID is not found that will be same as no match!!!
    """

    set_value = {
        'name': name,
        'comment': comment,
        'objType': 'email',
        'uuid': uuid,
        'smtpServer': smtpServer,
        'smtpPort': smtpPort,
        'from': from_email,
        'to': to_email
    }

    set_value = ibmsecurity.utilities.tools.json_sort(set_value)

    ret_obj = get_all(isamAppliance)
    for obj in ret_obj['data']['emailObjects']:
        if uuid is None and obj['name'] == name:
            return True
        elif ibmsecurity.utilities.tools.json_sort(obj) == set_value:
            return True

    return False


def compare(isamAppliance1, isamAppliance2):
    """
    Compare email objects between two appliances
    """
    ret_obj1 = get_all(isamAppliance1)
    ret_obj2 = get_all(isamAppliance2)

    for obj in ret_obj1['data']['emailObjects']:
        del obj['uuid']
    for obj in ret_obj2['data']['emailObjects']:
        del obj['uuid']

    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=['uuid'])



import logging

logger = logging.getLogger(__name__)

uri = "/mga/runtime_profile/cluster/v1"
requires_modules = ["mga", "federation"]
requires_version = "8.0.0.0"


def get(isamAppliance, check_mode=False, force=False):
    """
    Retrieving the status of the clustered runtime profiles
    """
    return isamAppliance.invoke_get("Retrieving the status of the clustered runtime profiles",
                                    f"{uri}",
                                    requires_modules=requires_modules, requires_version=requires_version)


def execute(isamAppliance, operation, check_mode=False, force=False):
    """
    Stopping, starting, restarting, or reloading the clustered runtime profiles

    :param isamAppliance:
    :param operation:
    :return:
    """

    if check_mode is True:
        return isamAppliance.create_return_object(changed=True)
    else:
        return isamAppliance.invoke_put("Stopping, starting, restarting, or reloading the clustered runtime profiles",
                                        f"{uri}",
                                        {
                                            'operation': operation
                                        },
                                        requires_modules=requires_modules,
                                        requires_version=requires_version)



import logging
import os.path
import re

from ibmsecurity.utilities.tools import files_same, get_random_temp_dir, version_compare
import shutil

logger = logging.getLogger(__name__)


def get_all(isamAppliance, kdb_id, check_mode=False, force=False):
    """
    Retrieving signer certificate names and details in a certificate database
    """
    return isamAppliance.invoke_get("Retrieving signer certificate names and details in a certificate database",
                                    f"/isam/ssl_certificates/{kdb_id}/signer_cert")


def get(isamAppliance, kdb_id, cert_id=None, label=None, check_mode=False, force=False):
    """
    Retrieving a signer certificate from a certificate database
    use cert_id (legacy) or label , mutually exclusive, label takes precedence
    """
    if label is None:
        retObj = isamAppliance.invoke_get("Retrieving a signer certificate from a certificate database",
                                    f"/isam/ssl_certificates/{kdb_id}/signer_cert/{cert_id}",
                                    ignore_error=True)
    else:
        cert_id = label
        retObj = isamAppliance.invoke_get("Retrieving a signer certificate from a certificate database",
                                    f"/isam/ssl_certificates/{kdb_id}/signer_cert/{label}",
                                    ignore_error=True)
    if retObj.get('rc', 0) == 404:
        return isamAppliance.create_return_object(rc=404, warnings=[f"{cert_id} does not exist in {kdb_id}"])
    else:
        return retObj

def load(isamAppliance, kdb_id, label, server, port, check_remote=False, check_mode=False, force=False):
    """
    Load a certificate from a server

    check_remote controls if ansible should check remote certificate by retrieving it or simply by
    checking for existence of the label in the kdb
    """
    if check_remote:
      logger.debug("Compare remote certificate with the one on the appliance. Use check_remote=False to switch to simple label checking.")
      tmp_check = _check_load(isamAppliance, kdb_id, label, server, port)
    else:
      logger.debug("Check for existence of the label in the kdb. Use check_remote=True to switch to advanced remote certificate with appliance certificate checking.")
      tmp_check = _check(isamAppliance, kdb_id, label)

    if force is True or tmp_check is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_post(
                "Load a certificate from a server",
                f"/isam/ssl_certificates/{kdb_id}/signer_cert",
                {
                    "operation": 'load',
                    "label": label,
                    "server": server,
                    "port": port
                })

    return isamAppliance.create_return_object()


def _check_expired(notafter_epoch):
    """
    Can be used to check for expired certs
    Returns True if expired, False otherwise
    """
    import time
    epoch_time = int(time.time())
    cert_epoch = int(notafter_epoch)
    return cert_epoch < epoch_time


def _check_load(isamAppliance, kdb_id, label, server, port):
    """
    Checks if certificate to be loaded on the Appliance exists and if so, whether it is different from
    the one on the remote host.

    If the certificate exists on the Appliance, but has a different label,
    we return True, so that load() takes no action.

    If the requested label matches an existing label on the appliance,
    but the certs are different, check to see if cert is expired.  If so, replace cert.
    If not, do not replace.
    """
    import ssl
    remote_cert_pem = ssl.get_server_certificate((server, port))

    # Look for remote_cert_pem on in the signer certs on the appliance
    ret_obj = get_all(isamAppliance, kdb_id)
    for cert_data in ret_obj['data']:
        cert_id = cert_data['id']
        cert_pem = get(isamAppliance, kdb_id, cert_id)['data']['contents']
        if cert_id == label:  # label exists on appliance already
            if cert_pem == remote_cert_pem:
                logger.debug("The certificate already exits on the appliance with the same label name and same content.")
                return True  # both the labels and certificates match
            else:
                # Labels match, but the certs are different, so we need to update it.
                # However, you cannot load a cert with the same label name onto the appliance, since you get
                #   CTGSK2021W A duplicate certificate already exists in the database.
                # We delete the cert from the appliance and return False to the load() function,
                # so that we can load the new one
                ret_obj = delete(isamAppliance, kdb_id, cert_id)
                logger.debug("Labels match, but the certs are different, so we need to update it.")
                return False
        else:
            if cert_pem == remote_cert_pem:  # cert on the appliance, but with a different name
                logger.info(
                    f"The certifcate is already on the appliance, but it has a different label name. The existing label name is {label} and requested label name is {cert_id}")
                return True
    return False


def delete(isamAppliance, kdb_id, cert_id=None, label=None, check_mode=False, force=False):
    """
    Deleting a signer certificate from a certificate database
    """
    if label is not None:
        cert_id = label

    if force or _check(isamAppliance, kdb_id, cert_id) is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            try:
                # Assume Python3 and import package
                from urllib.parse import quote
            except ImportError:
                # Now try to import Python2 package
                from urllib import quote

            # URL being encoded primarily to handle spaces and other special characers in them
            f_uri = f"/isam/ssl_certificates/{kdb_id}/signer_cert/{cert_id}"
            full_uri = quote(f_uri)
            return isamAppliance.invoke_delete(
                "Deleting a signer certificate from a certificate database", full_uri)

    return isamAppliance.create_return_object()


def export_cert(isamAppliance, kdb_id, cert_id, filename, check_mode=False, force=False):
    """
    Exporting a signer certificate from a certificate database
    """
    import os.path

    if force or _check(isamAppliance, kdb_id, cert_id):
        if check_mode is False:  # No point downloading a file if in check_mode
            return isamAppliance.invoke_get_file(
                "Export a certificate database",
                f"/isam/ssl_certificates/{kdb_id}/signer_cert/{cert_id}?export",
                filename)

    return isamAppliance.create_return_object()


def import_cert(isamAppliance, kdb_id, cert, label, preserve_label='false', check_mode=False, force=False):
    """
    Importing a signer certificate into a certificate database
    cert can be a file or a string
    """
    # Let's do some simple check
    # check if the string begins with -----BEGIN CERTIFICATE-----
    #
    if cert.startswith('-----BEGIN CERTIFICATE-----'):
      if force or _check_import_string(isamAppliance, kdb_id, label, cert, check_mode=check_mode):
          if check_mode:
              return isamAppliance.create_return_object(changed=True)
          else:
              json_data = {}
              json_data['label'] = label
              json_data['cert'] = cert
              json_data['operation'] = "import" # this is missing from the documentation.
              if version_compare(isamAppliance.facts['version'], "10.0.5.0") >= 0:
                  json_data['preserve_label'] = preserve_label.lower()
              ret_obj = isamAppliance.invoke_post("Create signer cert from certificate string",
                                               f"/isam/ssl_certificates/{kdb_id}/signer_cert",
                                               json_data,
                                               ignore_error=True)
              if preserve_label.lower() == 'true':
                  if ret_obj['rc'] == 500:
                      # this most likely means the certificate exists already (with a different name ?)
                      if "DPWAP0134E" in ret_obj['data'].get('message', ''):
                          return isamAppliance.create_return_object(rc=0, warnings=(f"Not updating existing certificate (preserve_label=true): {ret_obj.get('data', {}).get('message', '')}"), changed=False)
                      # this most likely means the certificate exists already (with a different name ?)
                      if "duplicate" in ret_obj['data'].get('message', ''):
                          return isamAppliance.create_return_object(rc=0, warnings=(f"Not updating existing certificate (preserve_label=true): {ret_obj.get('data', {}).get('message', '')}"), changed=False)
              else:
                  if ret_obj['rc'] == 500:
                      # this most likely means the certificate exists already (with a different name ?)
                      if "DPWAP0134E" in ret_obj['data'].get('message', ''):
                          return isamAppliance.create_return_object(rc=0, warnings=(f"Not updating existing certificate (preserve_label=false): {ret_obj.get('data', {}).get('message', '')}"), changed=False)
              if ret_obj['rc'] == 404:
                  logger.info(f"Not found... which is odd")
              return ret_obj

    else:
      if force or _check_import(isamAppliance, kdb_id, label, cert, check_mode=check_mode):
          if check_mode:
              return isamAppliance.create_return_object(changed=True)
          else:
              if version_compare(isamAppliance.facts['version'], "10.0.5.0") < 0:
                  return isamAppliance.invoke_post_files(
                      "Importing a signer certificate into a certificate database",
                      f"/isam/ssl_certificates/{kdb_id}/signer_cert",
                      [
                          {
                              'file_formfield': 'cert',
                              'filename': cert,
                              'mimetype': 'application/octet-stream'
                          }
                      ],
                      {'label': label})
              else:
                  # Preserve label true , throws 500 error
                  ret_obj = isamAppliance.invoke_post_files(
                      "Importing a signer certificate into a certificate database",
                      f"/isam/ssl_certificates/{kdb_id}/signer_cert",
                      [
                          {
                              'file_formfield': 'cert',
                              'filename': cert,
                              'mimetype': 'application/octet-stream'
                          }
                      ],
                      {'label': label,
                       'preserve_label': preserve_label.lower()},
                      ignore_error=True
                  )
                  if preserve_label.lower() == 'true':
                      if ret_obj['rc'] == 500:
                          # this most likely means the certificate exists already (with a different name ?)
                          if "DPWAP0134E" in ret_obj['data'].get('message', ''):
                              return isamAppliance.create_return_object(rc=0, warnings=(
                                  f"Not updating existing certificate (preserve_label=true): {ret_obj.get('data', {}).get('message', '')}"),
                                                                        changed=False)
                          # this most likely means the certificate exists already (with a different name ?)
                          if "duplicate" in ret_obj['data'].get('message', ''):
                              return isamAppliance.create_return_object(rc=0, warnings=(
                                  f"Not updating existing certificate (preserve_label=true): {ret_obj.get('data', {}).get('message', '')}"),
                                                                        changed=False)
                  else:
                      if ret_obj['rc'] == 500:
                          # this most likely means the certificate exists already (with a different name ?)
                          if "DPWAP0134E" in ret_obj['data'].get('message', ''):
                              return isamAppliance.create_return_object(rc=0, warnings=(
                                  f"Not updating existing certificate (preserve_label=false): {ret_obj.get('data', {}).get('message', '')}"),
                                                                        changed=False)
                  if ret_obj['rc'] == 404:
                      logger.info(f"Not found... which is odd")
                  return ret_obj

    return isamAppliance.create_return_object()


def _check(isamAppliance, kdb_id, cert_id):
    """
    Check if signer certificate already exists in certificate database
    """
    ret_obj = get_all(isamAppliance, kdb_id)

    for certdb in ret_obj['data']:
        if certdb.get('id', None) == cert_id:
            return True

    return False


def _check_import_string(isamAppliance, kdb_id, label, certstring, check_mode=False):
    cert_pem = get(isamAppliance, kdb_id, label)
    cert_pem = cert_pem.get("data", {})
    cert_pem = cert_pem.get("contents", "")
    if cert_pem == "":
        # No certificate found.  Fine.
        return True
    if cert_pem.replace(" ", "").replace("\n", "").replace("\r", "") == certstring.replace(" ", "").replace("\n", "").replace("\r", ""):
        logger.debug(f"Certificate already exists with same label {label}")
        return False
    return True


def _check_import(isamAppliance, kdb_id, cert_id, filename, check_mode=False):
    """
    Checks if certificate on the Appliance  exists and if so, whether it is different from
    the one stored in filename
    This is not a full check; because the certificate could exist under a different label !
    filename may be a chain, in that case we need to compare a bit more
    """
    with open(filename) as file:
        newcert = file.read() # tuple
        # split into separate certs , if it's a chain
        newcert = re.findall(r'(-----[BEGIN \S\ ]+?-----[\S\s]+?-----[END \S\ ]+?-----)', newcert)

    cert_pem = get(isamAppliance, kdb_id, cert_id)
    cert_pem = cert_pem.get("data", {})
    cert_pem = cert_pem.get("contents", "")
    if cert_pem == "":
        # No certificate found.  Fine.
        return True

    for nc in newcert:
        if cert_pem.replace(" ", "").replace("\n", "").replace("\r", "") == nc.replace(" ", "").replace("\n", "").replace("\r", ""):
            logger.debug(f"Signer Certificate already exists with same label {cert_id}")
            return False
    return True


def compare(isamAppliance1, isamAppliance2, kdb_id):
    """
    Compare signer certificates in certificate database between two appliances
    """
    ret_obj1 = get_all(isamAppliance1, kdb_id)
    ret_obj2 = get_all(isamAppliance2, kdb_id)

    import ibmsecurity.utilities.tools
    return ibmsecurity.utilities.tools.json_compare(ret_obj1, ret_obj2, deleted_keys=[])



import logging
import os

import ibmsecurity
from ibmsecurity.utilities import tools

logger = logging.getLogger(__name__)

uri = "/wga/apiac/resource/instance"
requires_modules = ["wga"]
requires_version = "9.0.7"


def get_all(isamAppliance, instance_name, check_mode=False, force=False):
    """
    Retrieving the list of all files in the API Access Control documentation root
    """
    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force or instance_exist:
        return isamAppliance.invoke_get(
            "Retrieving the list of all files in the API Access Control documentation root ",
            "{0}/{1}/documentation/".format(uri, instance_name),
            requires_modules=requires_modules, requires_version=requires_version)


def get(isamAppliance, instance_name, file_name, check_mode=False, force=False):
    """
    Retrieving the list of all files in the API Access Control documentation root
    """
    instance_exist, warnings = _check_instance_exist(isamAppliance, instance_name)

    if force or instance_exist:
        return isamAppliance.invoke_get(
            "Retrieving the list of all files in the API Access Control documentation root ",
            "{0}/{1}/documentation/{2}".format(uri, instance_name, file_name),
            requires_modules=requires_modules, requires_version=requires_version)


def add(isamAppliance, instance_name, type, file_name=None, dir_name=None, contents=None,
        check_mode=False, force=False):
    """
    Creating a file or directory in the API Access Control documentation
    """
    exist, warnings = _check_exist(isamAppliance, instance_name, file_name)

    if force is True or exist is False:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "type": type
            }
            if file_name != None:
                json_data['file_name'] = file_name

            if dir_name != None:
                json_data['dir_name'] = dir_name

            if contents != None:
                json_data['contents'] = contents

            return isamAppliance.invoke_post(
                "Creating a file or directory in the API Access Control documentation ",
                "{0}/{1}/documentation".format(uri, instance_name),
                json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def update(isamAppliance, instance_name, file_name, contents, type='file', check_mode=False, force=False):
    """
    Updating a file in the API Access Control documentation root
    """
    same_contents, warnings = _check_contents(isamAppliance, instance_name, file_name, contents)

    if force or not same_contents:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "type": type,
                "contents": contents
            }

            return isamAppliance.invoke_put(
                "Updating a file in the API Access Control documentation root  ",
                "{0}/{1}/documentation/{2}".format(uri, instance_name, file_name),
                json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def set(isamAppliance, instance_name, file_name, contents, type='file', check_mode=False, force=False):
    exist, warnings = _check_exist(isamAppliance, instance_name, file_name)

    if exist:
        return update(isamAppliance=isamAppliance, instance_name=instance_name, file_name=file_name, contents=contents,
                      type=type, check_mode=check_mode, force=force)
    else:
        return add(isamAppliance=isamAppliance, instance_name=instance_name, type=type, file_name=file_name,
                   contents=contents,
                   check_mode=check_mode, force=force)


def rename_directory(isamAppliance, instance_name, file_name, new_name, type='directory', check_mode=False,
                     force=False):
    """
    Renaming a directory in the API Access Control documentation root
    """
    exists, warnings = _check_exist(isamAppliance, instance_name, file_name)

    if force or exists:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "type": type,
                "new_name": new_name
            }

            return isamAppliance.invoke_put(
                "Renaming a directory in the API Access Control documentation root",
                "{0}/{1}/documentation/{2}".format(uri, instance_name, file_name),
                json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def rename_file(isamAppliance, instance_name, file_name, new_name, type='file', check_mode=False, force=False):
    """
    Renaming a file in the API Access Control documentation root
    """
    exists, warnings = _check_exist(isamAppliance, instance_name, file_name)

    if force or exists:
        if check_mode:
            return isamAppliance.create_return_object(changed=True)
        else:
            json_data = {
                "type": type,
                "new_name": new_name
            }

            return isamAppliance.invoke_put(
                "Renaming a file in the API Access Control documentation root",
                "{0}/{1}/documentation/{2}".format(uri, instance_name, file_name),
                json_data,
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def delete(isamAppliance, instance_name, file_name, check_mode=False, force=False):
    """
    Deleting a file or directory from the API Access Control
    """
    exists, warnings = _check_exist(isamAppliance, instance_name, file_name)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_delete(
                "Deleting a file or directory from the API Access Control",
                "{0}/{1}/documentation/{2}".format(uri, instance_name, file_name),
                requires_modules=requires_modules, requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def export_file(isamAppliance, instance_name, file_name, file_path, check_mode=False, force=False):
    """
    Exporting a file in the API Access Control documentation root
    """
    if os.path.exists(file_path) is True:
        warn_str = "File {0} already exists".format(file_path)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    exists, warnings = _check_exist(isamAppliance, instance_name, file_name)

    if force is True or exists is True:
        if check_mode is True:
            return isamAppliance.create_return_object(changed=True)
        else:
            return isamAppliance.invoke_get_file(
                "Exporting a file in the API Access Control documentation root",
                "{0}/{1}/documentation/{2}?export=true".format(uri, instance_name, file_name),
                file_path,
                requires_modules=requires_modules,
                requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def import_file(isamAppliance, instance_name, file_path, file_name="", check_mode=False, force=False):
    """
    Importing a file to the API Access Control documentation root
    file_path: location of the file to be uploaded.  for example: /home/user/file1.json
    file_name: name of the directory path and filename in API Documentation Root.  for example: dir/subdir or dir/file1
    """
    if os.path.exists(file_path) is False:
        warn_str = "File {0} already exists".format(file_path)
        warnings = [warn_str]
        return isamAppliance.create_return_object(warnings=warnings)

    same_contents, warnings = _check_contents(isamAppliance=isamAppliance, instance_name=instance_name,
                                              file_name=file_name, file=file_path)

    if force is True or same_contents is False:
        return isamAppliance.invoke_post_files("Importing a file to the API Access Control documentation root",
                                               "{0}/{1}/documentation/{2}?uiCalled=true".format(uri, instance_name,
                                                                                                file_name),
                                               [
                                                   {
                                                       'file_formfield': 'file',
                                                       'filename': file_path,
                                                       'mimetype': 'application/octet-stream'
                                                   }
                                               ],
                                               {
                                                   'type': 'file',
                                                   'force': True
                                               },
                                               requires_version=requires_version)

    return isamAppliance.create_return_object(warnings=warnings)


def _check_instance_exist(isamAppliance, instance_name):
    ret_obj = ibmsecurity.isam.web.api_access_control.resources.get_all_instances(isamAppliance)

    for obj in ret_obj['data']:
        if obj['name'] == instance_name:
            return True, ret_obj['warnings']

    return False, ret_obj['warnings']


def _check_exist(isamAppliance, instance_name, file_name):
    try:
        ret_obj = get(isamAppliance, instance_name, file_name)
        if ret_obj['data'] != []:
            return True, ret_obj['warnings']
        else:
            return False, ret_obj['warnings']

    except Exception as e:
        warnings = ["Exception: {0}".format(e)]
        return False, warnings


def _check_contents(isamAppliance, instance_name, file_name, contents=None, file=None):
    try:
        ret_obj = get(isamAppliance, instance_name, file_name)

        if contents != None:
            if ret_obj['data']['contents'] == contents:
                return True, ret_obj['warnings']
            else:
                return False, ret_obj['warnings']
        elif file != None:
            with open(file, 'rt') as myfile:
                new_contents = myfile.read()
            if ret_obj['data']['contents'] == new_contents:
                return True, ret_obj['warnings']
            else:
                return False, ret_obj['warnings']
        else:
            return False, ret_obj['warnings']

    except Exception as e:
        warnings = ["Exception occurred: {0}.".format(e)]
        return True, warnings


def compare(isamAppliance1, isamAppliance2, instance1_name, instance2_name=None):
    """
    Compare documentation root between two appliances
    """

    if instance2_name is None or instance2_name == '':
        instance2_name = instance1_name

    ret_obj1 = get_all(isamAppliance1, instance1_name)
    ret_obj2 = get_all(isamAppliance2, instance2_name)

    return tools.json_compare(ret_obj1, ret_obj2)
